/* +---------------------------------------------------------------------------+
   |          The Mobile Robot Programming Toolkit (MRPT) C++ library          |
   |                                                                           |
   |                   http://mrpt.sourceforge.net/                            |
   |                                                                           |
   |   Copyright (C) 2005-2009  University of Malaga                           |
   |                                                                           |
   |    This software was written by the Machine Perception and Intelligent    |
   |      Robotics Lab, University of Malaga (Spain).                          |
   |    Contact: Jose-Luis Blanco  <jlblanco@ctima.uma.es>                     |
   |                                                                           |
   |  This file is part of the MRPT project.                                   |
   |                                                                           |
   |     MRPT is free software: you can redistribute it and/or modify          |
   |     it under the terms of the GNU General Public License as published by  |
   |     the Free Software Foundation, either version 3 of the License, or     |
   |     (at your option) any later version.                                   |
   |                                                                           |
   |   MRPT is distributed in the hope that it will be useful,                 |
   |     but WITHOUT ANY WARRANTY; without even the implied warranty of        |
   |     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         |
   |     GNU General Public License for more details.                          |
   |                                                                           |
   |     You should have received a copy of the GNU General Public License     |
   |     along with MRPT.  If not, see <http://www.gnu.org/licenses/>.         |
   |                                                                           |
   +---------------------------------------------------------------------------+ */

#include <mrpt/precomp_core.h>  // Only for precomp. headers, include all libmrpt-core headers.


#include <mrpt/utils/CTicTac.h>
#include <mrpt/poses/CPosePDFGaussian.h>
#include <mrpt/poses/CPosePDFParticles.h>
#include <mrpt/slam/CActionCollection.h>
#include <mrpt/slam/CSensoryFrame.h>

#include <mrpt/random.h>
#include <mrpt/math_mrpt.h>

#include <set>

using namespace mrpt;
using namespace mrpt::bayes;
using namespace mrpt::poses;
using namespace mrpt::math;
using namespace mrpt::utils;
using namespace std;

IMPLEMENTS_SERIALIZABLE( CPosePDFParticles, CPosePDF, mrpt::poses )

/*---------------------------------------------------------------
	Constructor
  ---------------------------------------------------------------*/
CPosePDFParticles::CPosePDFParticles( size_t M ) :
	options(),
	m_pfAuxiliaryPFOptimal_estimatedProb()
{
	m_particles.resize(M);

	for (CParticleList::iterator it=m_particles.begin();it!=m_particles.end();it++)
		it->d = new CPose2D();

	static CPose2D	nullPose(0,0,0);
	resetDeterministic( nullPose );
}

/*---------------------------------------------------------------
						operator =
  ---------------------------------------------------------------*/
void  CPosePDFParticles::copyFrom(const CPosePDF &o)
{
	MRPT_TRY_START;

	CParticleList::iterator			itDest;
	CParticleList::const_iterator	itSrc;

	if (this == &o) return;		// It may be used sometimes

	if (o.GetRuntimeClass()==CLASS_ID(CPosePDFParticles))
	{
		const CPosePDFParticles	*pdf = static_cast<const CPosePDFParticles*>( &o );

		// Both are m_particles:
		if (m_particles.size()==pdf->m_particles.size())
		{
			for ( itSrc=pdf->m_particles.begin(), itDest = m_particles.begin();
				   itSrc!=pdf->m_particles.end();
			      itSrc++, itDest++ )
			{
				(*itDest->d) = (*itSrc->d);
				itDest->log_w = itSrc->log_w;
			}
		}
		else
		{
			for ( itDest = m_particles.begin();itDest!=m_particles.end();itDest++ )
				delete itDest->d;

			m_particles.resize( pdf->m_particles.size() );

			for ( itSrc=pdf->m_particles.begin(), itDest = m_particles.begin();
				   itSrc!=pdf->m_particles.end();
			      itSrc++, itDest++ )
			{
				itDest->d = new CPose2D( *itSrc->d );
				itDest->log_w = itSrc->log_w;
			}
		}
	}
	else
	if (o.GetRuntimeClass()==CLASS_ID(CPosePDFGaussian))
	{
		const CPosePDFGaussian	*pdf = static_cast<const CPosePDFGaussian*>( &o );
		size_t M = m_particles.size();
		std::vector<vector_double>			parts;
		std::vector<vector_double>::iterator partsIt;

		mrpt::random::randomNormalMultiDimensionalMany(pdf->cov,M,parts);

		clearParticles();
		m_particles.resize(M);

		for ( itDest = m_particles.begin(),partsIt=parts.begin();itDest!=m_particles.end();itDest++,partsIt++ )
		{
			itDest->log_w = 0;
            itDest->d = new CPose2D( ( pdf->mean.x + (*partsIt)[0] ),
									 ( pdf->mean.y + (*partsIt)[1] ),
                                     ( pdf->mean.phi + (*partsIt)[2] ) );
			itDest->d->normalizePhi();
		}

	}

	MRPT_TRY_END;
}

/*---------------------------------------------------------------
	Destructor
  ---------------------------------------------------------------*/
CPosePDFParticles::~CPosePDFParticles( )
{
	clear();
}

/*---------------------------------------------------------------
			clear
  ---------------------------------------------------------------*/
void  CPosePDFParticles::clear( )
{
	clearParticles();
}

/*---------------------------------------------------------------
						getEstimatedPose
  Returns an estimate of the pose, (the mean, or mathematical expectation of the PDF), computed
		as a weighted average over all m_particles.
 ---------------------------------------------------------------*/
CPose2D  CPosePDFParticles::getEstimatedPose() const
{
	CPose2D			est(0,0,0);
	CPose2D			p;
	size_t			i,n = m_particles.size();
	double			phi,w,W=0;
	double			W_phi_R=0,W_phi_L=0;
	double			phi_R=0,phi_L=0;


	if (!n) return est;

	// First: XY
	// -----------------------------------
	for (i=0;i<n;i++)
	{
		p  = *m_particles[i].d;
		w  = exp(m_particles[i].log_w);
		W += w;

		est.x+= p.x * w;
		est.y+= p.y * w;

		// PHI is special:
		phi = p.phi;
		if (fabs(phi)>1.5707963267948966192313216916398f)
		{
			// LEFT HALF: 0,2pi
			if (phi<0) phi = (M_2PI + phi);

			phi_L += phi * w;
			W_phi_L += w;
		}
		else
		{
			// RIGHT HALF: -pi,pi
			phi_R += phi * w;
			W_phi_R += w;
		}
	}

	est*=(1/W);

	// Next: PHI
	// -----------------------------------
	// The mean value from each side:
	if (W_phi_L>0)	phi_L /= W_phi_L;  // [0,2pi]
	if (W_phi_R>0)	phi_R /= W_phi_R;  // [-pi,pi]

	// Left side to [-pi,pi] again:
	if (phi_L>M_PI) phi_L = phi_L - M_2PI;

	// The total mean:
	est.phi =  ((phi_L * W_phi_L + phi_R * W_phi_R )/(W_phi_L+W_phi_R));

	return est;
}

/*---------------------------------------------------------------
						getEstimatedCovariance
  ---------------------------------------------------------------*/
CMatrixD  CPosePDFParticles::getEstimatedCovariance() const
{
	CMatrixD	cov(3,3);
	CPose2D		mean = getEstimatedPose();
	size_t		i,n = m_particles.size();
	double		var_x=0,var_y=0,var_p=0,var_xy=0,var_xp=0,var_yp=0;
	double		mean_phi = mean.phi;

	if (mean_phi<0) mean_phi = M_2PI + mean_phi;

	cov.zeros();

	double lin_w_sum = 0;

	for (i=0;i<n;i++) lin_w_sum+= exp( m_particles[i].log_w );
	if (lin_w_sum==0) lin_w_sum=1;

	for (i=0;i<n;i++)
	{
		double w = exp( m_particles[i].log_w ) / lin_w_sum;

		// Manage 1 PI range:
		double	err_x   = m_particles[i].d->x - mean.x;
		double	err_y   = m_particles[i].d->y - mean.y;
		double	err_phi = math::wrapToPi( fabs(m_particles[i].d->phi - mean_phi) );

		var_x+= square(err_x)*w;
		var_y+= square(err_y)*w;
		var_p+= square(err_phi)*w;
		var_xy+= err_x*err_y*w;
		var_xp+= err_x*err_phi*w;
		var_yp+= err_y*err_phi*w;
	}

	if (n<2)
	{
		// Not enought information to estimate the variance:
		cov.zeros(3,3);
	}
	else
	{
		// Unbiased estimation of variance:
		cov(0,0) = var_x;
		cov(1,1) = var_y;
		cov(2,2) = var_p;

		cov(1,0) = cov(0,1) = var_xy;
		cov(2,0) = cov(0,2) = var_xp;
		cov(1,2) = cov(2,1) = var_yp;

	}

	return cov;
}


/*---------------------------------------------------------------
						writeToStream
  ---------------------------------------------------------------*/
void  CPosePDFParticles::writeToStream(CStream &out,int *version) const
{
	if (version)
		*version = 0;
	else
	{
		writeParticlesToStream( out );
	}
}

/*---------------------------------------------------------------
						readFromStream
  ---------------------------------------------------------------*/
void  CPosePDFParticles::readFromStream(CStream &in, int version)
{
	switch(version)
	{
	case 0:
		{
			readParticlesFromStream( in );
		} break;
	default:
		MRPT_THROW_UNKNOWN_SERIALIZATION_VERSION(version)

	};
}

/*---------------------------------------------------------------

			prediction_and_update_pfStandardProposal

 ---------------------------------------------------------------*/
void  CPosePDFParticles::prediction_and_update_pfStandardProposal(
	const mrpt::slam::CActionCollection	* actions,
	const mrpt::slam::CSensoryFrame		* sf,
	const bayes::CParticleFilter::TParticleFilterOptions &PF_options )
{
	MRPT_TRY_START;

	size_t				i,M = m_particles.size();
	CActionRobotMovement2DPtr robotMovement;

	// ----------------------------------------------------------------------
	//						PREDICTION STAGE
	// ----------------------------------------------------------------------
	ASSERT_(actions!=NULL);

	// Find a robot movement estimation:
	robotMovement = actions->getBestMovementEstimation();

	// Assure one has been found:
	if ( !robotMovement )
		THROW_EXCEPTION("Action list does not contain any CActionRobotMovement2D derived object!");

	//  Prediction:
	// -----------------------------
	if ( !PF_options.adaptiveSampleSize )
	{
		// ---------------------
		//   FIXES SAMPLE SIZE
		// ---------------------
		// Initialize random sample generator:
		mrpt::poses::CPoseRandomSampler	poseSamplesGen;
		poseSamplesGen.setPosePDF( robotMovement->poseChange );

		CPose2D		increment_i;
		//robotMovement->poseChange->drawManySamples( M, rndSamples );	// Generate gaussian-distributed 2D-pose increments according to mean-cov:


		for (i=0;i<M;i++)			// Update particle poses:
		{
			poseSamplesGen.drawSample(increment_i);

			// Add pose increment.
			(*m_particles[i].d) = (*m_particles[i].d) + increment_i;
		}
	} // end of fixed sample size
	else
	{
		// -------------------------------------------------------------
		//   ADAPTIVE SAMPLE SIZE
		// Implementation of Dieter Fox's KLD algorithm
		//  JLBC (31/OCT/2006)
		// TODO: Optimize the bins search (i.e. tree structures?)
		// -------------------------------------------------------------
        // Create a (x,y,phi) bins storage:
		std::set<TPoseBin,lt_TPoseBin>						poseBins;
		std::set<TPoseBin,lt_TPoseBin>::const_iterator		it;
		TPoseBin											p;

		// The new particle set:
		std::deque<CPose2D*>			newParticles;
		size_t							N, Nx = options.KLD_params.KLD_minSampleSize;
		CPose2D							increment_i, *newPose;
		double							delta_1 = 1.0 - options.KLD_params.KLD_delta;
		double							epsilon_1 = 0.5 / options.KLD_params.KLD_epsilon;

		// Prepare data for executing "fastDrawSample"
		prepareFastDrawSample(PF_options);

		// -------------------------------------------
        //		THE MAIN DRAW SAMPLING LOOP
		// -------------------------------------------
		do
		{
			// Draw a robot movement increment:
			//robotMovement->poseChange->drawSingleSample( increment_i );
			robotMovement->drawSingleSample( increment_i );

			// generate the new particle:
			newPose = new CPose2D( *m_particles[ fastDrawSample(PF_options) ].d );
			*newPose = *newPose + increment_i;

			// Add to the new-m_particles list:
			newParticles.push_back( newPose );

			// Now, look if the particle falls in a new bin or not:
			// --------------------------------------------------------
			p.x = round( newPose->x / options.KLD_params.KLD_binSize_XY );
			p.y = round( newPose->y / options.KLD_params.KLD_binSize_XY );
			p.phi = round( newPose->phi / options.KLD_params.KLD_binSize_PHI );

			it = poseBins.find( p );
			if (it==poseBins.end())
			{
				// It falls into a new bin:
				// Add to the poseBins:
				poseBins.insert( p );

				// K = K + 1
				size_t	K = poseBins.size();
				if ( K>1) //&& newParticles.size() > options.KLD_minSampleSize )
				{
					// Update the number of m_particles!!
					Nx =  round(epsilon_1 * math::chiSquareQuantile(delta_1,K-1));
					//printf("k=%u \tn=%u \tNx:%u\n", k, newParticles.size(), Nx);
				}
			}

			N = newParticles.size();

		} while (	N < max(Nx,(size_t)options.KLD_params.KLD_minSampleSize) &&
					N < options.KLD_params.KLD_maxSampleSize );


		// Substitute old by new particle set:
		// -------------------------------------------------
		clear();	// Free old m_particles memory:
		m_particles.resize(N);
		CParticleList::iterator	itDest;
		std::deque<CPose2D*>::iterator			itSrc;
		for (itDest=m_particles.begin(),itSrc=newParticles.begin();itDest!=m_particles.end();itDest++,itSrc++)
		{
			itDest->d = *itSrc;
			itDest->log_w = 0;
		}
		newParticles.clear();

	} // end of ADAPTIVE SAMPLE SIZE

	// ----------------------------------------------------------------------
	//						UPDATE STAGE
	// ----------------------------------------------------------------------
	M = m_particles.size();
	if (sf!=NULL)
	{
		// A map MUST be supplied!
		ASSERT_(options.metricMap || options.metricMaps.size()>0);
		if (!options.metricMap)
			ASSERT_(options.metricMaps.size() == M);

		// Update particle's likelihood using the particle's pose:
		CParticleList::iterator	it;
		for (it=m_particles.begin(),i=0;it!=m_particles.end();it++,i++)
			it->log_w += auxiliarComputeObservationLikelihood(PF_options, this,i,sf, it->d ) * PF_options.powFactor;
	};

	MRPT_TRY_END;
}

/*---------------------------------------------------------------
			particlesEvaluator_AuxPFOptimal
 ---------------------------------------------------------------*/
double  CPosePDFParticles::particlesEvaluator_AuxPFOptimal(
	const bayes::CParticleFilter::TParticleFilterOptions &PF_options,
	const CParticleFilterCapable	*obj,
	size_t index,
	const void * action,
	const void * observation )
{
	MRPT_TRY_START;

	// Compute the quantity:
	//     w[i]·p(zt|z^{t-1},x^{[i],t-1},X)
	// See paper: [blanco2007...]

	// Take the previous particle weight:
	// --------------------------------------------
	//CPosePDFParticles *pdf = const_cast<CPosePDFParticles*> ( static_cast<const CPosePDFParticles*>(obj) );
	const CPosePDFParticles *pdf = static_cast<const CPosePDFParticles*>(obj);

	double		ret = pdf->m_particles[index].log_w;

	// , take the mean of the posterior density:
	// --------------------------------------------
	THROW_EXCEPTION("FIX THIS!!");
	CPose2D		x_predict( *pdf->m_particles[index].d + *static_cast<const CPose2D*>(action) );

	// and compute the obs. likelihood:
	// --------------------------------------------
	return ret + ( pdf->m_pfAuxiliaryPFOptimal_estimatedProb[index] = auxiliarComputeObservationLikelihood(
		PF_options,
		pdf,
		index,
		static_cast<const CSensoryFrame*>(observation),
		&x_predict ) );

	MRPT_TRY_END;
}

/*---------------------------------------------------------------
			particlesEvaluator_AuxPFStandard
 ---------------------------------------------------------------*/
double  CPosePDFParticles::particlesEvaluator_AuxPFStandard(
	const bayes::CParticleFilter::TParticleFilterOptions &PF_options,
	const CParticleFilterCapable	*obj,
	size_t index,
	const void * action,
	const void * observation )
{
	MRPT_TRY_START;

	// Compute w[i]·p(z_t | mu_t^i)
	//   with mu_t^i being the mean of the new robot pose

	// Take the previous particle weight:
	// --------------------------------------------
	const CPosePDFParticles *pdf = static_cast<const CPosePDFParticles*>(obj);

	double		ret = pdf->m_particles[index].log_w;

	// , take the mean of the posterior density:
	// --------------------------------------------
	CPose2D		x_predict( *pdf->m_particles[index].d + *static_cast<const CPose2D*>(action) );

	// and compute the obs. likelihood:
	// --------------------------------------------
	return ret + ( pdf->m_pfAuxiliaryPFStandard_estimatedProb[index] = auxiliarComputeObservationLikelihood(
		PF_options,
		pdf,
		index,
		static_cast<const CSensoryFrame*>(observation),
		&x_predict ) );

	MRPT_TRY_END;
}


/*---------------------------------------------------------------

			prediction_and_update_pfAuxiliaryPFOptimal

 ---------------------------------------------------------------*/
void  CPosePDFParticles::prediction_and_update_pfAuxiliaryPFOptimal(
	const mrpt::slam::CActionCollection	* actions,
	const mrpt::slam::CSensoryFrame		* sf,
	const bayes::CParticleFilter::TParticleFilterOptions &PF_options )
{
	MRPT_TRY_START;

	size_t						i,k,N,M = m_particles.size();
	CActionRobotMovement2DPtr robotMovement;
	//std::vector<vector_double>	rndSamples;

	// ----------------------------------------------------------------------
	//							PRELIMINAR CHECKS:
	// ----------------------------------------------------------------------
	ASSERT_(sf!=NULL);
	ASSERT_(actions!=NULL);

	robotMovement = actions->getBestMovementEstimation();	// Find a robot movement estimation:
	if ( !robotMovement )									// Assure one has been found:
		THROW_EXCEPTION("Action list does not contain any CActionRobotMovement2D derived object!");

	ASSERT_(options.metricMap || options.metricMaps.size()>0);
	if (!options.metricMap)
	{
		ASSERT_(options.metricMaps.size() == M);
	}


	// ----------------------------------------------------------------------
	//		0) Common part:  Prepare m_particles "draw" and compute
	// ----------------------------------------------------------------------
	m_movementDrawer.setPosePDF( robotMovement->poseChange );

	m_pfAuxiliaryPFOptimal_estimatedProb.resize(M);
//	m_maxLikelihood.clear();
//	m_maxLikelihood.resize(M,0);
//	m_movementDrawMaximumLikelihood.resize(M);

	// Prepare data for executing "fastDrawSample"
	CTicTac		tictac;
	printf("[prepareFastDrawSample] Computing...");
	tictac.Tic();
	prepareFastDrawSample(
		PF_options,
		particlesEvaluator_AuxPFOptimal,
		NULL, //robotMovement, // Not used
		sf );
	printf("Done! in %.06f ms\n",tictac.Tac()*1e3f);

#if 1  	/** DEBUG **/
	printf("[prepareFastDrawSample] max      (log) = %10.06f\n",  math::maximum(m_pfAuxiliaryPFOptimal_estimatedProb) );
	printf("[prepareFastDrawSample] max-mean (log) = %10.06f\n", -math::mean(m_pfAuxiliaryPFOptimal_estimatedProb) + math::maximum(m_pfAuxiliaryPFOptimal_estimatedProb) );
	printf("[prepareFastDrawSample] max-min  (log) = %10.06f\n", -math::minimum(m_pfAuxiliaryPFOptimal_estimatedProb) + math::maximum(m_pfAuxiliaryPFOptimal_estimatedProb) );
#endif	/*****/

	// Now we have the vector "m_fastDrawProbability" filled out with:
	//     w[i]·p(zt|z^{t-1},x^{[i],t-1},X)
	//  where X is the robot pose prior (as implemented in
	//  the aux. function "particlesEvaluator_AuxPFOptimal"),

	// Precompute a list of "random" samples from the movement model:
/*	std::vector<CPose2D>			movementDraws;
	movementDraws.resize(PF_options.pfAuxFilterOptimal_MaximumSearchSamples*3);
	for (std::vector<CPose2D>::iterator	it = movementDraws.begin(); it!=movementDraws.end(); it++)
		robotMovement->poseChange->drawSingleSample( *it );
*/

	std::deque<CPose2D*>			newParticles;
	vector_double					newParticlesWeight;

	// We need the (aproximate) maximum likelihood value for each
	//  previous particle [i]:
	//
	//     max{ p( z^t | data^[i], x_(t-1)^[i], u_(t) ) }
	//
	// We store "-1" if the max. value has not been computed yet:
	vector_double					maxLikelihood(M, -1e300 );
	CPose2D							movementDraw,newPose,oldPose;
	double							acceptanceProb,newPoseLikelihood,ratioLikLik;
	unsigned int					timeoutCount, MAX_TIMEOUT = 1000;
	double							bestNewPoseAccProb;
	CPose2D							bestNewPose;

	if ( !PF_options.adaptiveSampleSize )
	{
		// ----------------------------------------------------------------------
		//						1) FIXED SAMPLE SIZE VERSION
		// ----------------------------------------------------------------------
		newParticles.resize(M);
		newParticlesWeight.resize(M);


		for (i=0;i<M;i++)
		{
			// Generate a new particle:
			//   (a) Draw a "t-1" m_particles' index:
			// ----------------------------------------------------------------
			k = fastDrawSample( PF_options );
			oldPose = *m_particles[k].d;

            //   (b) Rejection-sampling: Draw a new robot pose from x[k],
			//       and accept it with probability p(zk|x) / maxLikelihood:
			// ----------------------------------------------------------------
			// Is the maxLikelihood already computed?
			double			maxLik_k = maxLikelihood[k];
			CPose2D			maxLikNewPose, drawnSample;

			if (maxLik_k<0)
			{
				// NO -> Compute now and save it:
				for (unsigned int q=0 ; q<PF_options.pfAuxFilterOptimal_MaximumSearchSamples; q++)
				{
					// Draw new robot pose:
					newPose = oldPose + m_movementDrawer.drawSample(drawnSample); //movementDraws[q];
					double lik = auxiliarComputeObservationLikelihood(PF_options, this,k,sf,&newPose);
					if (lik>maxLik_k)
					{
						maxLik_k = lik;
						maxLikNewPose = newPose;
					}
				}
				maxLikelihood[k] = maxLik_k;
			}

			// Rejection-sampling:
			timeoutCount = 0;
			bestNewPoseAccProb = 0;
			//static unsigned int	q = PF_options.pfAuxFilterOptimal_MaximumSearchSamples;
			do
			{
				// Draw new robot pose:
				//robotMovement->drawSingleSample( movementDraw );
				//newPose = oldPose + movementDraw;
				newPose = oldPose + m_movementDrawer.drawSample(drawnSample); //movementDraws[(q++) % movementDraws.size()];

				// Compute acceptance probability:
				newPoseLikelihood = auxiliarComputeObservationLikelihood(PF_options,this,k,sf,&newPose);
				ratioLikLik = exp(newPoseLikelihood - maxLikelihood[k]);
				acceptanceProb = min( 1.0, ratioLikLik );

				// Save the best "newPose" for the case of having a "timeout":
				if (acceptanceProb>bestNewPoseAccProb)
				{
					bestNewPoseAccProb	= acceptanceProb;
					bestNewPose			= newPose;
				}

                if ( ratioLikLik > 1)
				{
					if (ratioLikLik>1.2)
					{
						//  DEBUG
						//printf("\n[pfAuxiliaryPFOptimal] Warning!! p(z|x)/p(z|x*)=%f\n",ratioLikLik);
					}
					maxLikelihood[k] = newPoseLikelihood; //  :'-( !!!
				}

			} while ( acceptanceProb < mrpt::random::RandomUni(0.0,0.999) && (++timeoutCount)<MAX_TIMEOUT );

			if (timeoutCount>=MAX_TIMEOUT)
				newPose = bestNewPose;

			// Insert the new particle!:
			newParticles[i] = new CPose2D( newPose );
			// And its weight:
			newParticlesWeight[i] = 1; //newPoseLikelihood / m_pfAuxiliaryPFOptimal_estimatedProb[k];
		} // for i

	} // end fixed sample size
	else
	{
		// ----------------------------------------------------------------------
		//	 				2) ADAPTIVE SAMPLE SIZE VERSION
		//
		// Implementation of Dieter Fox's KLD algorithm
		//  JLBC (3/OCT/2006)
		// TODO: Optimize the bins search (i.e. tree structures?)
		// -------------------------------------------------------------
        // Create a (x,y,phi) bins storage:
		std::set<TPoseBin,lt_TPoseBin>						poseBins;
		std::set<TPoseBin,lt_TPoseBin>::const_iterator		it;
		TPoseBin											p;
		CPose2D  drawnSample;

		// The new particle set:
		newParticles.clear();
		newParticlesWeight.clear();
		unsigned int					Nx = options.KLD_params.KLD_minSampleSize;
//			CPose2D							increment_i, *newPose;
		double							delta_1 = 1.0 - options.KLD_params.KLD_delta;
		double							epsilon_1 = 0.5 / options.KLD_params.KLD_epsilon;
		k = 0;

		// -------------------------------------------
        //	  THE MAIN KLD-BASED DRAW SAMPLING LOOP
		// -------------------------------------------
		do
		{
			// Generate a new particle:
			//   (a) Draw a "t-1" m_particles' index:
			// ----------------------------------------------------------------
			k = fastDrawSample( PF_options );
			oldPose = *m_particles[k].d;

            //   (b) Rejection-sampling: Draw a new robot pose from x[k],
			//       and accept it with probability p(zk|x) / maxLikelihood:
			// ----------------------------------------------------------------
			// Is the maxLikelihood already computed?
			double		maxLik_k = maxLikelihood[k];
			if (maxLik_k<0)
			{
				// NO -> Compute now and save it:
				for (unsigned int q=0 ; q<PF_options.pfAuxFilterOptimal_MaximumSearchSamples; q++)
				{
					// Draw new robot pose:
					//robotMovement->poseChange->drawSingleSample( movementDraw );
					newPose = oldPose + m_movementDrawer.drawSample(drawnSample); //movementDraws[q];
					double lik = auxiliarComputeObservationLikelihood(PF_options,this,k,sf,&newPose);
					maxLik_k = max(maxLik_k,lik);
				}
				maxLikelihood[k] = maxLik_k;
			}

			// Rejection-sampling:
			timeoutCount = 0;
			bestNewPoseAccProb = 0;
			do
			{
				// Draw new robot pose:
				//robotMovement->poseChange->drawSingleSample( movementDraw );
				robotMovement->drawSingleSample( movementDraw );
				newPose = oldPose + movementDraw;

				// Compute acceptance probability:
				newPoseLikelihood = auxiliarComputeObservationLikelihood(PF_options,this,k,sf,&newPose);
				ratioLikLik = exp( newPoseLikelihood - maxLikelihood[k]);
				acceptanceProb = min( 1.0, ratioLikLik );

				// Save the best "newPose" for the case of having a "timeout":
				if (acceptanceProb>bestNewPoseAccProb)
				{
					bestNewPoseAccProb	= acceptanceProb;
					bestNewPose			= newPose;
				}

                if ( ratioLikLik > 1)
				{
					if (ratioLikLik>1.2)
					{
						/** DEBUG **/
						//printf("\n[pfAuxiliaryPFOptimal] Warning!! p(z|x)/p(z|x*)=%f\n",ratioLikLik);
						/**/
					}
					maxLikelihood[k] = newPoseLikelihood; //  :'-( !!!
				}

			} while ( acceptanceProb < mrpt::random::RandomUni(0.0,0.999) && (++timeoutCount)<MAX_TIMEOUT );
//			} while ( acceptanceProb < 0.95  && (++timeoutCount)<MAX_TIMEOUT );

			if (timeoutCount>=MAX_TIMEOUT)
				newPose = bestNewPose;

			// Insert the new particle!:
			newParticles.push_back( new CPose2D( newPose ) );
			// And its weight:
			newParticlesWeight.push_back( 1 ); //newPoseLikelihood / m_pfAuxiliaryPFOptimal_estimatedProb[k] );

			// Now, look if the particle falls in a new bin or not:
			// --------------------------------------------------------
			p.x = round( newPose.x / options.KLD_params.KLD_binSize_XY );
			p.y = round( newPose.y / options.KLD_params.KLD_binSize_XY );
			p.phi = round( newPose.phi / options.KLD_params.KLD_binSize_PHI );

			it = poseBins.find( p );
			if (it==poseBins.end())
			{
				// It falls into a new bin:
				// Add to the poseBins:
				poseBins.insert( p );

				// K = K + 1
				size_t	K = poseBins.size();
				if ( K>1) //&& newParticles.size() > options.KLD_params.KLD_minSampleSize )
				{
					// Update the number of m_particles!!
					Nx =  round(epsilon_1 * math::chiSquareQuantile(delta_1,K-1));
					//printf("k=%u \tn=%u \tNx:%u\n", k, newParticles.size(), Nx);
				}
			}

			N = newParticles.size();

		} while (	N < options.KLD_params.KLD_maxSampleSize &&
					N < max(Nx,options.KLD_params.KLD_minSampleSize) );

	} // end adaptive sample size

	// Substitute old by new particle set:
	// -------------------------------------------------
	N = newParticles.size();
	clear();	// Free old m_particles memory:
	m_particles.resize(N);

	CParticleList::iterator	itDest;
	std::deque<CPose2D*>::iterator			itSrc;
	vector_double::iterator					itW;

	for (itDest=m_particles.begin(), itSrc = newParticles.begin(), itW = newParticlesWeight.begin();
			itDest!=m_particles.end();
		    itDest++,itSrc++,itW++)
	{
		itDest->d = *itSrc;
		itDest->log_w = *itW;
	}
	newParticles.clear();
	normalizeWeights();

	MRPT_TRY_END;
}


/*---------------------------------------------------------------
							resetDeterministic
	Reset PDF to a single point and set the number of m_particles
 ---------------------------------------------------------------*/
void  CPosePDFParticles::resetDeterministic( const CPose2D &location,
										size_t	particlesCount)
{
	CParticleList::iterator		it;

	if (particlesCount>0)
	{
		clear();
		m_particles.resize(particlesCount);
		for (it=m_particles.begin();it!=m_particles.end();it++)
			it->d = new CPose2D();
	}

	for (it=m_particles.begin();it!=m_particles.end();it++)
	{
		*it->d	= location;
		it->log_w	= 0;
	}
}

/*---------------------------------------------------------------
						resetUniform
 ---------------------------------------------------------------*/
void  CPosePDFParticles::resetUniform(
	const double & x_min,
	const double & x_max,
	const double & y_min,
	const double & y_max,
	const double & phi_min,
	const double & phi_max,
	const int	&particlesCount)
{
	MRPT_TRY_START;

	if (particlesCount>0)
	{
		clear();
		m_particles.resize(particlesCount);
		for (int i=0;i<particlesCount;i++)
			m_particles[i].d = new CPose2D();
	}

	size_t		i,M = m_particles.size();
	for (i=0;i<M;i++)
	{
		m_particles[i].d->x   = mrpt::random::RandomUni( x_min, x_max );
		m_particles[i].d->y   = mrpt::random::RandomUni( y_min, y_max );
		m_particles[i].d->phi = mrpt::random::RandomUni( phi_min, phi_max );
		m_particles[i].log_w=0;
	}

	MRPT_TRY_END;
}

/*---------------------------------------------------------------
						resetUniformFreeSpace
 ---------------------------------------------------------------*/
void  CPosePDFParticles::resetUniformFreeSpace(
	COccupancyGridMap2D		*theMap,
	const double &					freeCellsThreshold,
	const int	 &					particlesCount,
	const double &					x_min,
	const double &					x_max ,
	const double &					y_min,
	const double &					y_max,
	const double &					phi_min ,
	const double &					phi_max)
{
	MRPT_TRY_START;

	if (particlesCount>0)
	{
		clear();
		m_particles.resize(particlesCount);
		for (int i=0;i<particlesCount;i++)
			m_particles[i].d = new CPose2D();
	}

	size_t				i,M = m_particles.size();
	int					sizeX = theMap->getSizeX();
	int					sizeY = theMap->getSizeY();
	double				gridRes = theMap->getResolution();
	std::vector<double>	freeCells_x,freeCells_y;
	size_t				nFreeCells;
	unsigned int		xIdx1,xIdx2;
	unsigned int		yIdx1,yIdx2;

	freeCells_x.reserve( sizeX * sizeY );
	freeCells_y.reserve( sizeX * sizeY );

	if (x_min>theMap->getXMin())
			xIdx1 = max(0, theMap->x2idx( x_min ) );
	else	xIdx1 = 0;
	if (x_max<theMap->getXMax())
			xIdx2 = min(sizeX-1, theMap->x2idx( x_max ) );
	else	xIdx2 = sizeX-1;
	if (y_min>theMap->getYMin())
			yIdx1 = max(0, theMap->y2idx( y_min ) );
	else	yIdx1 = 0;
	if (y_max<theMap->getYMax())
			yIdx2 = min(sizeY-1, theMap->y2idx( y_max ) );
	else	yIdx2 = sizeY-1;


	for (unsigned int x=xIdx1;x<=xIdx2;x++)
		for (unsigned int y=yIdx1;y<=yIdx2;y++)
			if (theMap->getCell(x,y)>=freeCellsThreshold)
			{
				freeCells_x.push_back(theMap->idx2x(x));
				freeCells_y.push_back(theMap->idx2y(y));
			}

	nFreeCells = freeCells_x.size();

	// Assure that map is not fully occupied!
	ASSERT_( nFreeCells );

	// Generate pose m_particles:
	for (i=0;i<M;i++)
	{
		int idx = round(mrpt::random::RandomUni(0,nFreeCells-1.05f));

		m_particles[i].d->x   = freeCells_x[idx] + mrpt::random::RandomUni( -gridRes, gridRes );
		m_particles[i].d->y   = freeCells_y[idx] + mrpt::random::RandomUni( -gridRes, gridRes );
		m_particles[i].d->phi = mrpt::random::RandomUni( phi_min, phi_max );
		m_particles[i].log_w=0;
	}

	MRPT_TRY_END;
}

/*---------------------------------------------------------------
						saveToTextFile
   Save PDF's m_particles to a text file. In each line it
      will go: "x y phi weight"
 ---------------------------------------------------------------*/
void  CPosePDFParticles::saveToTextFile(const std::string &file) const
{
	FILE	*f=os::fopen(file.c_str(),"wt");
	if (!f) return;

	for (unsigned int i=0;i<m_particles.size();i++)
		os::fprintf(f,"%f %f %f %e\n",
				m_particles[i].d->x,
				m_particles[i].d->y,
				m_particles[i].d->phi,
				m_particles[i].log_w );

	os::fclose(f);
}

/*---------------------------------------------------------------
						getParticlePose
 ---------------------------------------------------------------*/
CPose2D	 CPosePDFParticles::getParticlePose(size_t i) const
{
	return *m_particles[i].d;
}

/*---------------------------------------------------------------
						changeCoordinatesReference
 ---------------------------------------------------------------*/
void  CPosePDFParticles::changeCoordinatesReference(const CPose2D &newReferenceBase )
{
	CParticleList::iterator	it;

	for (it=m_particles.begin();it!=m_particles.end();it++)
		(*it->d) = newReferenceBase + (*it->d);
}

/*---------------------------------------------------------------
					drawSingleSample
 ---------------------------------------------------------------*/
void  CPosePDFParticles::drawSingleSample( CPose2D &outPart ) const
{
	double									uni = mrpt::random::RandomUni(0.0f,0.9999f);
	double									cum = 0;
	CParticleList::const_iterator	it;

	for (it=m_particles.begin();it!=m_particles.end();it++)
	{
		cum+= exp(it->log_w);
		if ( uni<= cum )
		{
			outPart= (*it->d);
			return;
		}
	}

	// Might not come here normally:
	outPart = *(m_particles.end()-1)->d;
}

/*---------------------------------------------------------------
					drawManySamples
 ---------------------------------------------------------------*/
void  CPosePDFParticles::drawManySamples(
	size_t						N,
	std::vector<vector_double>	&outSamples ) const
{
	CPose2D		*ptr;

	CParticleFilter::TParticleFilterOptions  PF_options;
	PF_options.adaptiveSampleSize = true;
	PF_options.resamplingMethod = CParticleFilter::prMultinomial;

	prepareFastDrawSample( PF_options );

	outSamples.resize(N);
	for (size_t i=0;i<N;i++)
	{
		ptr = m_particles[ fastDrawSample( PF_options ) ].d;

		// Copy pose:
		outSamples[i].resize(3);
		outSamples[i][0] = ptr->x;
		outSamples[i][1] = ptr->y;
		outSamples[i][2] = ptr->phi;
	}
}


/*---------------------------------------------------------------
						+=
 ---------------------------------------------------------------*/
void  CPosePDFParticles::operator += ( CPose2D Ap)
{
	CParticleList::iterator	it;

	for (it=m_particles.begin();it!=m_particles.end();it++)
		(*it->d) = (*it->d) + Ap;
}

/*---------------------------------------------------------------
					append
 ---------------------------------------------------------------*/
void  CPosePDFParticles::append( CPosePDFParticles &o )
{
	for (unsigned int i=0;i<o.m_particles.size();i++)
	{
		CParticleData		part;
		part.d = new CPose2D( *o.m_particles[i].d );
		part.log_w = o.m_particles[i].log_w;
		m_particles.push_back( part );
	}

	normalizeWeights();
}

/*---------------------------------------------------------------
					inverse
 ---------------------------------------------------------------*/
void	 CPosePDFParticles::inverse(CPosePDF  &o) const
{
	MRPT_TRY_START;
	ASSERT_( o.GetRuntimeClass() == CLASS_ID(CPosePDFParticles) );
	CPosePDFParticles	*out = static_cast<CPosePDFParticles*>( &o );


	out->copyFrom( *this );
	static CPose2D		nullPose(0,0,0);

	for (unsigned int i=0;i<out->m_particles.size();i++)
		(*out->m_particles[i].d) = nullPose - (*out->m_particles[i].d);

	MRPT_TRY_END;
}

/*---------------------------------------------------------------
					getMostLikelyParticle
 ---------------------------------------------------------------*/
CPose2D	 CPosePDFParticles::getMostLikelyParticle() const
{
	CParticleList::const_iterator	it, itMax=m_particles.begin();
	double		max_w = -1e300;


	for (it=m_particles.begin();it!=m_particles.end();it++)
	{
		if (it->log_w > max_w)
		{
			itMax = it;
			max_w = it->log_w;
		}
	}

	return *itMax->d;
}

/*---------------------------------------------------------------
					bayesianFusion
 ---------------------------------------------------------------*/
void  CPosePDFParticles::bayesianFusion( CPosePDF &p1, CPosePDF &p2, const double &minMahalanobisDistToDrop  )
{
	MRPT_UNUSED_PARAM(p1);MRPT_UNUSED_PARAM(p2);MRPT_UNUSED_PARAM(minMahalanobisDistToDrop);

	THROW_EXCEPTION("Not implemented yet!");
}

/*---------------------------------------------------------------
					evaluatePDF_parzen
 ---------------------------------------------------------------*/
double  CPosePDFParticles::evaluatePDF_parzen(
	const double &	x,
	const double &	y,
	const double &	phi,
	const double &	stdXY,
	const double &	stdPhi ) const
{
	double	difPhi, ret = 0;

	for (CParticleList::const_iterator	it=m_particles.begin();it!=m_particles.end();it++)
	{
		difPhi = math::wrapToPi( phi-it->d->phi );

		ret += exp(it->log_w) *
			   math::normalPDF( sqrt( square(x-it->d->x) + square(y-it->d->y)) , 0, stdXY ) *
			   math::normalPDF( fabs( difPhi ), 0, stdPhi );
	}

	return ret;
}

/*---------------------------------------------------------------
					saveParzenPDFToTextFile
 ---------------------------------------------------------------*/
void  CPosePDFParticles::saveParzenPDFToTextFile(
	const char	*fileName,
	const double &		x_min,
	const double &		x_max,
	const double &		y_min,
	const double &		y_max,
	const double &		phi,
	const double &		stepSizeXY,
	const double &		stdXY,
	const double &		stdPhi ) const
{
	FILE	*f=os::fopen(fileName,"wt");
	if (!f) return;

	for (double y=y_min; y<y_max; y+=stepSizeXY)
	{
		for (double x=x_min; x<x_max; x+=stepSizeXY)
		{
			os::fprintf(f,"%f ",
				evaluatePDF_parzen(x,y,phi,stdXY,stdPhi) );
		} // y
		os::fprintf(f,"\n");
	} // x

	os::fclose(f);
}


/*---------------------------------------------------------------
					TPredictionParams
 ---------------------------------------------------------------*/
CPosePDFParticles::TPredictionParams::TPredictionParams() :
	metricMap(NULL),
	metricMaps(),
	KLD_params()
{
}

/*---------------------------------------------------------------
				auxiliarComputeObservationLikelihood
 ---------------------------------------------------------------*/
double  CPosePDFParticles::auxiliarComputeObservationLikelihood(
	const bayes::CParticleFilter::TParticleFilterOptions &PF_options,
	const CParticleFilterCapable	*obj,
	size_t			particleIndexForMap,
	const CSensoryFrame	*observation,
	const CPose2D			*x )
{
	double			ret = 1;
	CMetricMap		*map;	// The map:

	const CPosePDFParticles *pdf = static_cast<const CPosePDFParticles*> (obj);

	if (pdf->options.metricMap)
			map = pdf->options.metricMap;
	else
	{
		ASSERT_( pdf->options.metricMaps.size()>particleIndexForMap );
		map = pdf->options.metricMaps[particleIndexForMap]; //->m_gridMaps[0];
	}

	// For each observation:
	for (CSensoryFrame::const_iterator it=observation->begin();it!=observation->end();it++)
		ret += map->computeObservationLikelihood( it->pointer(), *x );	// Compute the likelihood:

	// Done!
	return ret;
}


/*---------------------------------------------------------------

			prediction_and_update_pfAuxiliaryPFStandard

 ---------------------------------------------------------------*/
void  CPosePDFParticles::prediction_and_update_pfAuxiliaryPFStandard(
	const mrpt::slam::CActionCollection	* actions,
	const mrpt::slam::CSensoryFrame		* sf,
	const bayes::CParticleFilter::TParticleFilterOptions &PF_options )
{
	MRPT_TRY_START;

/*	size_t						i,k,N,M = m_particles.size();
	CActionRobotMovement2DPtr robotMovement;
	std::vector<vector_double>	rndSamples;

	// ----------------------------------------------------------------------
	//							PRELIMINAR CHECKS:
	// ----------------------------------------------------------------------
	ASSERT_(sf!=NULL);
	ASSERT_(actions!=NULL);

	robotMovement = actions->getBestMovementEstimation();	// Find a robot movement estimation:
	if ( !robotMovement )									// Assure one has been found:
		THROW_EXCEPTION("Action list does not contain any CActionRobotMovement2D derived object!");

	ASSERT_(options.metricMap || options.metricMaps.size()>0);
	if (!options.metricMap)
	{
		ASSERT_(options.metricMaps.size() == M);
	}

	// ----------------------------------------------------------------------
	//		0) Common part:  Prepare m_particles "draw" and compute
	// ----------------------------------------------------------------------
	// Pass the "mean" robot movement to the "weights" computing function:
	CPose2D		meanRobotMovement( robotMovement->poseChange->getEstimatedPose() );

	m_pfAuxiliaryPFOptimal_estimatedProb.resize(M);

	// Prepare data for executing "fastDrawSample"
	prepareFastDrawSample(
		PF_options,
		particlesEvaluator_AuxPFOptimal,
		&meanRobotMovement,
		sf );

	// Now we have the vector "m_fastDrawProbability" filled out with:
	//     w[i]·p(zt|z^{t-1},x^{[i],t-1},X)
	//  where X is the MEAN of the robot pose prior (as implemented in
	//  the aux. function "particlesEvaluator_AuxPFOptimal").

	// Precompute a list of "random" samples from the movement model:
	std::vector<CPose2D>			movementDraws;
	movementDraws.resize(PF_options.pfAuxFilterOptimal_MaximumSearchSamples*3);
	for (std::vector<CPose2D>::iterator	it = movementDraws.begin(); it!=movementDraws.end(); it++)
		robotMovement->poseChange->drawSingleSample( *it );

	std::deque<CPose2D*>			newParticles;
	vector_double					newParticlesWeight;

	// We need the (aproximate) maximum likelihood value for each
	//  previous particle [i]:
	//
	//     max{ p( z^t | data^[i], x_(t-1)^[i], u_(t) ) }
	//
	// We store "-1" if the max. value has not been computed yet:
	vector_double					maxLikelihood(M, -1e300 );
	CPose2D							movementDraw,newPose,oldPose;
	double							acceptanceProb,newPoseLikelihood,ratioLikLik;
	unsigned int					timeoutCount, MAX_TIMEOUT = 1000;
	double							bestNewPoseAccProb;
	CPose2D							bestNewPose;

	if ( !PF_options.adaptiveSampleSize )
	{
		// ----------------------------------------------------------------------
		//						1) FIXED SAMPLE SIZE VERSION
		// ----------------------------------------------------------------------
		newParticles.resize(M);
		newParticlesWeight.resize(M);


		for (i=0;i<M;i++)
		{
			// Generate a new particle:
			//   (a) Draw a "t-1" m_particles' index:
			// ----------------------------------------------------------------
			k = fastDrawSample( PF_options );
			oldPose = *m_particles[k].d;

            //   (b) Rejection-sampling: Draw a new robot pose from x[k],
			//       and accept it with probability p(zk|x) / maxLikelihood:
			// ----------------------------------------------------------------
			// Is the maxLikelihood already computed?
			double			maxLik_k = maxLikelihood[k];
			CPose2D			maxLikNewPose;

			if (maxLik_k<0)
			{
				// NO -> Compute now and save it:
				for (unsigned int q=0 ; q<PF_options.pfAuxFilterOptimal_MaximumSearchSamples; q++)
				{
					// Draw new robot pose:
					newPose = oldPose + movementDraws[q];
					double lik = auxiliarComputeObservationLikelihood(PF_options, this,k,sf,&newPose);
					if (lik>maxLik_k)
					{
						maxLik_k = lik;
						maxLikNewPose = newPose;
					}
				}
				maxLikelihood[k] = maxLik_k;
			}

			// Rejection-sampling:
			timeoutCount = 0;
			bestNewPoseAccProb = 0;
			static unsigned int	q = PF_options.pfAuxFilterOptimal_MaximumSearchSamples;
			do
			{
				// Draw new robot pose:
				//robotMovement->drawSingleSample( movementDraw );
				//newPose = oldPose + movementDraw;
				newPose = oldPose + movementDraws[(q++) % movementDraws.size()];

				// Compute acceptance probability:
				newPoseLikelihood = auxiliarComputeObservationLikelihood(PF_options,this,k,sf,&newPose);
				ratioLikLik = exp(newPoseLikelihood - maxLikelihood[k]);
				acceptanceProb = min( 1.0, ratioLikLik );

				// Save the best "newPose" for the case of having a "timeout":
				if (acceptanceProb>bestNewPoseAccProb)
				{
					bestNewPoseAccProb	= acceptanceProb;
					bestNewPose			= newPose;
				}

                if ( ratioLikLik > 1)
				{
					if (ratioLikLik>1.2)
					{
						//  DEBUG
						//printf("\n[pfAuxiliaryPFOptimal] Warning!! p(z|x)/p(z|x*)=%f\n",ratioLikLik);
					}
					maxLikelihood[k] = newPoseLikelihood; //  :'-( !!!
				}

			} while ( acceptanceProb < mrpt::random::RandomUni(0.0,0.999) && (++timeoutCount)<MAX_TIMEOUT );

			if (timeoutCount>=MAX_TIMEOUT)
				newPose = bestNewPose;

			// Insert the new particle!:
			newParticles[i] = new CPose2D( newPose );
			// And its weight:
			newParticlesWeight[i] = 1; //newPoseLikelihood / m_pfAuxiliaryPFOptimal_estimatedProb[k];
		} // for i

	} // end fixed sample size
	else
	{
		// ----------------------------------------------------------------------
		//	 				2) ADAPTIVE SAMPLE SIZE VERSION
		//
		// Implementation of Dieter Fox's KLD algorithm
		//  JLBC (3/OCT/2006)
		// TODO: Optimize the bins search (i.e. tree structures?)
		// -------------------------------------------------------------
        // Create a (x,y,phi) bins storage:
		std::set<TPoseBin,lt_TPoseBin>						poseBins;
		std::set<TPoseBin,lt_TPoseBin>::const_iterator		it;
		TPoseBin											p;

		// The new particle set:
		newParticles.clear();
		newParticlesWeight.clear();
		unsigned int					Nx = options.KLD_params.KLD_minSampleSize;
//			CPose2D							increment_i, *newPose;
		double							delta_1 = 1.0 - options.KLD_params.KLD_delta;
		double							epsilon_1 = 0.5 / options.KLD_params.KLD_epsilon;
		k = 0;

		// -------------------------------------------
        //	  THE MAIN KLD-BASED DRAW SAMPLING LOOP
		// -------------------------------------------
		do
		{
			// Generate a new particle:
			//   (a) Draw a "t-1" m_particles' index:
			// ----------------------------------------------------------------
			k = fastDrawSample( PF_options );
			oldPose = *m_particles[k].d;

            //   (b) Rejection-sampling: Draw a new robot pose from x[k],
			//       and accept it with probability p(zk|x) / maxLikelihood:
			// ----------------------------------------------------------------
			// Is the maxLikelihood already computed?
			double		maxLik_k = maxLikelihood[k];
			if (maxLik_k<0)
			{
				// NO -> Compute now and save it:
				for (unsigned int q=0 ; q<PF_options.pfAuxFilterOptimal_MaximumSearchSamples; q++)
				{
					// Draw new robot pose:
					//robotMovement->poseChange->drawSingleSample( movementDraw );
					newPose = oldPose + movementDraws[q];
					double lik = auxiliarComputeObservationLikelihood(PF_options,this,k,sf,&newPose);
					maxLik_k = max(maxLik_k,lik);
				}
				maxLikelihood[k] = maxLik_k;
			}

			// Rejection-sampling:
			timeoutCount = 0;
			bestNewPoseAccProb = 0;
			do
			{
				// Draw new robot pose:
				//robotMovement->poseChange->drawSingleSample( movementDraw );
				robotMovement->drawSingleSample( movementDraw );
				newPose = oldPose + movementDraw;

				// Compute acceptance probability:
				newPoseLikelihood = auxiliarComputeObservationLikelihood(PF_options,this,k,sf,&newPose);
				ratioLikLik = exp( newPoseLikelihood - maxLikelihood[k]);
				acceptanceProb = min( 1.0, ratioLikLik );

				// Save the best "newPose" for the case of having a "timeout":
				if (acceptanceProb>bestNewPoseAccProb)
				{
					bestNewPoseAccProb	= acceptanceProb;
					bestNewPose			= newPose;
				}

                if ( ratioLikLik > 1)
				{
					if (ratioLikLik>1.2)
					{
						//printf("\n[pfAuxiliaryPFOptimal] Warning!! p(z|x)/p(z|x*)=%f\n",ratioLikLik);
					}
					maxLikelihood[k] = newPoseLikelihood; //  :'-( !!!
				}

			} while ( acceptanceProb < mrpt::random::RandomUni(0.0,0.999) && (++timeoutCount)<MAX_TIMEOUT );
//			} while ( acceptanceProb < 0.95  && (++timeoutCount)<MAX_TIMEOUT );

			if (timeoutCount>=MAX_TIMEOUT)
				newPose = bestNewPose;

			// Insert the new particle!:
			newParticles.push_back( new CPose2D( newPose ) );
			// And its weight:
			newParticlesWeight.push_back( 1 ); //newPoseLikelihood / m_pfAuxiliaryPFOptimal_estimatedProb[k] );

			// Now, look if the particle falls in a new bin or not:
			// --------------------------------------------------------
			p.x = round( newPose.x / options.KLD_params.KLD_binSize_XY );
			p.y = round( newPose.y / options.KLD_params.KLD_binSize_XY );
			p.phi = round( newPose.phi / options.KLD_params.KLD_binSize_PHI );

			it = poseBins.find( p );
			if (it==poseBins.end())
			{
				// It falls into a new bin:
				// Add to the poseBins:
				poseBins.insert( p );

				// K = K + 1
				size_t	K = poseBins.size();
				if ( K>1) //&& newParticles.size() > options.KLD_params.KLD_minSampleSize )
				{
					// Update the number of m_particles!!
					Nx =  round(epsilon_1 * math::chiSquareQuantile(delta_1,K-1));
					//printf("k=%u \tn=%u \tNx:%u\n", k, newParticles.size(), Nx);
				}
			}

			N = newParticles.size();

		} while (	N < options.KLD_params.KLD_maxSampleSize &&
					N < max(Nx,options.KLD_params.KLD_minSampleSize) );

	} // end adaptive sample size

	// Substitute old by new particle set:
	// -------------------------------------------------
	N = newParticles.size();
	clear();	// Free old m_particles memory:
	m_particles.resize(N);

	CParticleList::iterator	itDest;
	std::deque<CPose2D*>::iterator			itSrc;
	vector_double::iterator					itW;

	for (itDest=m_particles.begin(), itSrc = newParticles.begin(), itW = newParticlesWeight.begin();
			itDest!=m_particles.end();
		    itDest++,itSrc++,itW++)
	{
		itDest->d = *itSrc;
		itDest->log_w = *itW;
	}
	newParticles.clear();
	normalizeWeights();
*/

	MRPT_TRY_END;
}



