/* +---------------------------------------------------------------------------+
   |          The Mobile Robot Programming Toolkit (MRPT) C++ library          |
   |                                                                           |
   |                   http://mrpt.sourceforge.net/                            |
   |                                                                           |
   |   Copyright (C) 2005-2009  University of Malaga                           |
   |                                                                           |
   |    This software was written by the Machine Perception and Intelligent    |
   |      Robotics Lab, University of Malaga (Spain).                          |
   |    Contact: Jose-Luis Blanco  <jlblanco@ctima.uma.es>                     |
   |                                                                           |
   |  This file is part of the MRPT project.                                   |
   |                                                                           |
   |     MRPT is free software: you can redistribute it and/or modify          |
   |     it under the terms of the GNU General Public License as published by  |
   |     the Free Software Foundation, either version 3 of the License, or     |
   |     (at your option) any later version.                                   |
   |                                                                           |
   |   MRPT is distributed in the hope that it will be useful,                 |
   |     but WITHOUT ANY WARRANTY; without even the implied warranty of        |
   |     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         |
   |     GNU General Public License for more details.                          |
   |                                                                           |
   |     You should have received a copy of the GNU General Public License     |
   |     along with MRPT.  If not, see <http://www.gnu.org/licenses/>.         |
   |                                                                           |
   +---------------------------------------------------------------------------+ */

#include <mrpt/precomp_core.h>  // Only for precomp. headers, include all libmrpt-core headers.


#include <mrpt/slam/CRangeBearingKFSLAM.h>
#include <mrpt/poses/CPosePDF.h>
#include <mrpt/poses/CPosePDFGaussian.h>
#include <mrpt/slam/CActionRobotMovement3D.h>

#include <mrpt/math/utils.h>
#include <mrpt/math/CMatrix.h>
#include <mrpt/math/CMatrixD.h>
#include <mrpt/utils/CTicTac.h>

using namespace mrpt::slam; using namespace mrpt::utils; using namespace mrpt::poses;
using namespace mrpt::utils;
using namespace std;

/*---------------------------------------------------------------
							Constructor
  ---------------------------------------------------------------*/
CRangeBearingKFSLAM::CRangeBearingKFSLAM( ) :
	options(),
	m_action(),m_SF(),
	m_IDs(),
	m_IDs_inverse(),
	mapPartitioner(),
	m_SFs(),
	m_lastPartitionSet()
{
	// -----------------------
	// INIT KF STATE
	m_xkk.resize(6,0);	// State: 6D pose (x,y,z,yaw,pitch,roll)

	// Initial cov:  Very small diagonal
	m_pkk.setSize(6,6);
	m_pkk.zeros();
	// -----------------------


	// Use SF-based matching (faster & easier for bearing-range observations with ID).
	mapPartitioner.options.useMapMatching  = false;

}

/*---------------------------------------------------------------
							Destructor
  ---------------------------------------------------------------*/
CRangeBearingKFSLAM::~CRangeBearingKFSLAM()
{

}

/*---------------------------------------------------------------
							getCurrentRobotPose
  ---------------------------------------------------------------*/
void CRangeBearingKFSLAM::getCurrentRobotPose(
	CPose3DPDFGaussian      &out_robotPose ) const
{
	MRPT_TRY_START

	ASSERT_(m_xkk.size()>=6);

	// Set 6D pose mean:
	out_robotPose.mean.setFromValues(
		m_xkk[0],
		m_xkk[1],
		m_xkk[2],
		m_xkk[3],
		m_xkk[4],
		m_xkk[5] );
	// and cov:
	CMatrixTemplateNumeric<KFTYPE>  COV(6,6);
	m_pkk.extractMatrix(0,0,COV);
	out_robotPose.cov = COV;

	MRPT_TRY_END
}


/*---------------------------------------------------------------
							getCurrentState
  ---------------------------------------------------------------*/
void CRangeBearingKFSLAM::getCurrentState(
	CPose3DPDFGaussian      &out_robotPose,
	std::vector<CPoint3D>  &out_landmarksPositions,
	std::map<unsigned int,CLandmark::TLandmarkID> &out_landmarkIDs,
	CVectorDouble      &out_fullState,
	CMatrixDouble      &out_fullCovariance ) const
{
	MRPT_TRY_START

	ASSERT_(m_xkk.size()>=6);

	// Set 6D pose mean:
	out_robotPose.mean.setFromValues(
		m_xkk[0],
		m_xkk[1],
		m_xkk[2],
		m_xkk[3],
		m_xkk[4],
		m_xkk[5] );
	// and cov:
	CMatrixTemplateNumeric<KFTYPE>  COV(6,6);
	m_pkk.extractMatrix(0,0,COV);
	out_robotPose.cov = COV;

	// Landmarks:
	ASSERT_( ((m_xkk.size() - 6) % 3)==0 );
	size_t i,nLMs = (m_xkk.size() - 6) / 3;
	out_landmarksPositions.resize(nLMs);
	for (i=0;i<nLMs;i++)
	{
		out_landmarksPositions[i].x = m_xkk[6+i*3+0];
		out_landmarksPositions[i].y = m_xkk[6+i*3+1];
		out_landmarksPositions[i].z = m_xkk[6+i*3+2];
	} // end for i

	// IDs:
	out_landmarkIDs = m_IDs_inverse;

    // Full state:
    out_fullState.resize( m_xkk.size() );
    for (i=0;i<m_xkk.size();i++)
		out_fullState[i] = m_xkk[i];

	// Full cov:
	out_fullCovariance = m_pkk;

	MRPT_TRY_END
}


/*---------------------------------------------------------------
						processActionObservation
  ---------------------------------------------------------------*/
void  CRangeBearingKFSLAM::processActionObservation(
    const CActionCollection	&action,
    const CSensoryFrame		&SF )
{
	MRPT_TRY_START

	m_action = &action;
	m_SF = &SF;

	// Sanity check:
	ASSERT_( m_IDs.size() == (m_pkk.getColCount() - get_vehicle_size() )/get_feature_size() );
	ASSERT_( m_IDs.size() == m_IDs_inverse.size() );

	// ===================================================================================================================
	// Here's the meat!: Call the main method for the KF algorithm, which will call all the callback methods as required:
	// ===================================================================================================================
	runOneKalmanIteration();


	// =============================================================
	//  ADD TO SFs SEQUENCE
	// =============================================================
    CPose3DPDFGaussian	auxPosePDF;
	getCurrentRobotPose( auxPosePDF );
    m_SFs.insert( &auxPosePDF , SF );

	// =============================================================
	//  UPDATE THE PARTITION GRAPH EXPERIMENT
	// =============================================================
	if (options.doPartitioningExperiment)
	{
	    if (options.partitioningMethod==0)
	    {
	        // Use spectral-graph technique:
            mapPartitioner.addMapFrame( SF, auxPosePDF );

            vector<vector_uint>   partitions;
            mapPartitioner.updatePartitions( partitions );
            m_lastPartitionSet = partitions;
	    }
	    else
	    {
	        // Fixed partitions every K observations:
	        vector<vector_uint>   partitions;
	        vector_uint           tmpCluster;

	        ASSERT_(options.partitioningMethod>1);
	        size_t    K = options.partitioningMethod;

	        for (size_t i=0;i<m_SFs.size();i++)
	        {
	            tmpCluster.push_back(i);
	            if ( (i % K) == 0 )
	            {
	                partitions.push_back(tmpCluster);
	                tmpCluster.clear();
	                tmpCluster.push_back(i);  // This observation "i" is shared between both clusters
	            }
	        }
            m_lastPartitionSet = partitions;
	    }

		printf("Partitions: %u\n", static_cast<unsigned>(m_lastPartitionSet.size() ));
	}

    MRPT_TRY_END
}



/** Must return the action vector u.
  * \param out_u The action vector which will be passed to OnTransitionModel
  */
void CRangeBearingKFSLAM::OnGetAction( CVectorTemplate<KFTYPE> &u )
{
	u.resize(6);

	// Get odometry estimation:
	CActionRobotMovement2DPtr actMov2D = m_action->getBestMovementEstimation();
	CActionRobotMovement3DPtr actMov3D = m_action->getActionByClass<CActionRobotMovement3D>();
	if (actMov3D)
	{
		u[0]=actMov3D->poseChange.mean.x;
		u[1]=actMov3D->poseChange.mean.y;
		u[2]=actMov3D->poseChange.mean.z;
		u[3]=actMov3D->poseChange.mean.yaw;
		u[4]=actMov3D->poseChange.mean.pitch;
		u[5]=actMov3D->poseChange.mean.roll;
	}
	else
	if (actMov2D)
	{
		CPose2D estMovMean( actMov2D->poseChange->getEstimatedPose() );
		u[0]=estMovMean.x;
		u[1]=estMovMean.y;
		u[2]=0;
		u[3]=estMovMean.phi;
		u[4]=0;
		u[5]=0;
	}
	else
	{
		// Left u as zeros
		for (size_t i=0;i<6;i++) u[i]=0;
	}

}


/** This virtual function musts implement the prediction model of the Kalman filter.
 */
void  CRangeBearingKFSLAM::OnTransitionModel(
	const CVectorTemplate<KFTYPE> &u,
	CVectorTemplate<KFTYPE>        &xv,
	bool &out_skipPrediction )
{
	MRPT_TRY_START

	// Do not update the vehicle pose & its covariance until we have some landmakrs in the map,
	// otherwise, we are imposing a lower bound to the best uncertainty from now on:
	if (m_xkk.size() == get_vehicle_size() )
	{
		out_skipPrediction = true;
		return;
	}

	ASSERT_(xv.size()>=6);
	ASSERT_(u.size()==6); // 6D odometry

	CPose3D  robotPose(xv[0],xv[1],xv[2],xv[3],xv[4],xv[5]);
	CPose3D  odoIncrement(u[0],u[1],u[2],u[3],u[4],u[5]);

	// Pose composition:
	robotPose = robotPose + odoIncrement;

	xv[0]=robotPose.x;
	xv[1]=robotPose.y;
	xv[2]=robotPose.z;
	xv[3]=robotPose.yaw;
	xv[4]=robotPose.pitch;
	xv[5]=robotPose.roll;

	MRPT_TRY_END
}

/** This virtual function musts calculate the Jacobian F of the prediction model.
 */
void  CRangeBearingKFSLAM::OnTransitionJacobian(
	CMatrixTemplateNumeric<KFTYPE> & F )
{
	MRPT_TRY_START

	// The jacobian of the transition function:  dfv_dxv
	CActionRobotMovement2DPtr act2D = m_action->getBestMovementEstimation();
	CActionRobotMovement3DPtr act3D = m_action->getActionByClass<CActionRobotMovement3D>();

	if (act3D && act2D)  THROW_EXCEPTION("Both 2D & 3D odometry are present!?!?")

	CPoint3D  Ap;

	if (act3D)
	{
		Ap = CPoint3D(act3D->poseChange.mean);
	}
	else if (act2D)
	{
		Ap = CPoint3D(act2D->poseChange->getEstimatedPose());
	}
	else
	{
		// No odometry at all:
		F .unit(6); // Unit diagonal
		return;
	}

	KFTYPE cy = cos(m_xkk[3]);	KFTYPE sy = sin(m_xkk[3]);
	KFTYPE cp = cos(m_xkk[4]);	KFTYPE sp = sin(m_xkk[4]);
	KFTYPE cr = cos(m_xkk[5]);	KFTYPE sr = sin(m_xkk[5]);

	KFTYPE xu = Ap.x;
	KFTYPE yu = Ap.y;
	KFTYPE zu = Ap.z;

	F .unit(6); // Unit diagonal

	F (0,3) = -sy*cp*xu+(-sy*sp*sr-cy*cr)*yu+(-sy*sp*cr+cy*sr)*zu;
	F (0,4) = -cy*sp*xu + cy*cp*sr*yu+cy*cp*cr*zu;
	F (0,5) = (cy*sp*cr+sy*sr)*yu+(-cy*sp*sr+sy*cr)*zu;
	F (1,3) = cy*cp*xu+(cy*sp*sr-sy*cr)*yu+(cy*sp*cr+sy*sr)*zu;
	F (1,4) = -sy*sp*xu+sy*cp*sr*yu+sy*cp*cr*zu;
	F (1,5) = (sy*sp*cr-cy*sr)*yu+(-sy*sp*sr-cy*cr)*zu;
	F (2,4) = -cp*xu-sp*sr*yu-sp*cr*zu;
	F (2,5) = cp*cr*yu-cp*sr*zu;

	MRPT_TRY_END
}

/** This virtual function musts calculate de noise matrix of the prediction model.
 */
void  CRangeBearingKFSLAM::OnTransitionNoise( CMatrixTemplateNumeric<KFTYPE> & Q )
{
	MRPT_TRY_START

	// The uncertainty of the 2D odometry, projected from the current position:
	CActionRobotMovement2DPtr act2D = m_action->getBestMovementEstimation();
	CActionRobotMovement3DPtr act3D = m_action->getActionByClass<CActionRobotMovement3D>();

	if (act3D && act2D)  THROW_EXCEPTION("Both 2D & 3D odometry are present!?!?")

	CPose3DPDFGaussian odoIncr;

	if (!act3D && !act2D)
	{
		// Use constant Q:
		Q.zeros(6,6);
		ASSERT_(options.stds_Q_no_odo.size()==Q.getColCount())

		for (size_t i=0;i<6;i++)
			Q(i,i) = square( options.stds_Q_no_odo[i]);
		return;
	}
	else
	{
		if (act2D)
		{
			// 2D odometry:
			CPosePDFGaussian odoIncr2D;
			odoIncr2D.copyFrom( *act2D->poseChange );
			odoIncr = odoIncr2D;

			// Avoid null values in the diagonal!
			odoIncr.cov(2,2)=        // z
			odoIncr.cov(4,4)=        // pitch
			odoIncr.cov(5,5)=1e-20f; // roll
		}
		else
		{
			// 3D odometry:
			odoIncr = act3D->poseChange;
		}
	}

	odoIncr.cov(2,2) +=  square(options.std_odo_z_additional);

	odoIncr.changeCoordinatesReference( CPose3D( m_xkk[0],m_xkk[1],m_xkk[2],m_xkk[3],m_xkk[4],m_xkk[5] ) );

	Q = odoIncr.cov;

	MRPT_TRY_END
}


/** Implements the observation prediction \f$ h_i(x) \f$ and the Jacobians \f$ \frac{\partial h_i}{\partial x} \f$ and (when applicable) \f$ \frac{\partial h_i}{\partial y_i} \f$.
  * \param in_z This is the same matrix returned by OnGetObservations(), passed here for reference.
  * \param in_data_association The vector returned by OnGetObservations(), passed here for reference.
  * \param in_full If set to true, all the Jacobians and predictions must be computed at once. Otherwise, just those for the observation in_obsIdx.
  * \param in_obsIdx If in_full=false, the row of the observation (in in_z and in_data_association) whose innovation & Jacobians are to be returned now.
  * \param out_innov The difference between the expected observation and the real one: \f$ \tilde{y} = z - h(x) \f$. It must be a vector of length O*N (in_full=true) or O (in_full=false), with O=get_observation_size() and N=number of rows in in_z.
  * \param out_Hx One or a vertical stack of \f$ \frac{\partial h_i}{\partial x} \f$.
  * \param out_Hy An empty matrix, or one or a vertical stack of \f$ \frac{\partial h_i}{\partial y_i} \f$.
  *
  *  out_Hx must consist of 1 (in_full=false) or N (in_full=true) vertically stacked \f$O \times V\f$ matrices, and out_Hy must consist of 1 or N \f$O \times F\f$ matrices, with:
  *  - N: number of rows in in_z (number of observed features, or 1 in general).
  *  - O: get_observation_size()
  *  - V: get_vehicle_size()
  *  - F: get_feature_size()
  */
void CRangeBearingKFSLAM::OnObservationModelAndJacobians(
	const CMatrixTemplateNumeric<KFTYPE> &in_z,
	const vector_int                     &data_association,
	const bool                           &in_full,
	const int                            &in_obsIdx,
	CVectorTemplate<KFTYPE>       		 &ytilde,
	CMatrixTemplateNumeric<KFTYPE>       &Hx,
	CMatrixTemplateNumeric<KFTYPE>       &Hy )
{
	MRPT_TRY_START

	// Get the sensor pose relative to the robot:
	CObservationBearingRangePtr obs = m_SF->getObservationByClass<CObservationBearingRange>();
	ASSERT_(obs);
	CPose3D   sensorPoseOnRobot( obs->sensorLocationOnRobot );

	/* -------------------------------------------
       Equations, obtained using matlab, of the relative 3D position of a landmark (xi,yi,zi), relative
          to a robot 6D pose (x0,y0,z0,y,p,r)
        Refer to technical report "6D EKF derivation...", 2008

        x0 y0 z0 y p r         % Robot's 6D pose
        x0s y0s z0s ys ps rs   % Sensor's 6D pose relative to robot
        xi yi zi % Absolute 3D landmark coordinates:

        Hx : dh_dxv   -> Jacobian of the observation model wrt the robot pose
        Hy : dh_dyi   -> Jacobian of the observation model wrt each landmark mean position

        Sizes:
	     h:  Lx3
         Hx: 3Lx6
         Hy: 3Lx3

          L=# of landmarks in the map (ALL OF THEM)
	  ------------------------------------------- */
	size_t i,nLMs = (m_xkk.size() - get_vehicle_size() ) / get_feature_size();


	// Mean of the prior of the robot pose:
	CPose3D  robotPose( m_xkk[0],m_xkk[1],m_xkk[2],m_xkk[3],m_xkk[4],m_xkk[5] );

	// Requested to build just one LM?
	if (!in_full)
        nLMs = 1;

    size_t  vehicle_size = get_vehicle_size();
    size_t  obs_size  = get_observation_size();
    size_t  feature_size = get_feature_size();

	ytilde.resize( obs_size*nLMs,0);
    Hx.setSize    ( obs_size*nLMs, vehicle_size );
    Hy.setSize    ( obs_size*nLMs, feature_size );

    // Robot 6D pose:
    KFTYPE x0 = m_xkk[0];
    KFTYPE y0 = m_xkk[1];
    KFTYPE z0 = m_xkk[2];
    KFTYPE y  = m_xkk[3];
    KFTYPE p  = m_xkk[4];
    KFTYPE r  = m_xkk[5];

    KFTYPE cy = cos(y); KFTYPE sy = sin(y);
    KFTYPE cp = cos(p); KFTYPE sp = sin(p);
    KFTYPE cr = cos(r); KFTYPE sr = sin(r);

    // Sensor 6D pose on robot:
    KFTYPE x0s = sensorPoseOnRobot.x;
    KFTYPE y0s = sensorPoseOnRobot.y;
    KFTYPE z0s = sensorPoseOnRobot.z;
    KFTYPE ys = sensorPoseOnRobot.yaw;
    KFTYPE ps = sensorPoseOnRobot.pitch;
    KFTYPE rs = sensorPoseOnRobot.roll;

    KFTYPE cys = cos(ys); KFTYPE sys = sin(ys);
    KFTYPE cps = cos(ps); KFTYPE sps = sin(ps);
    KFTYPE crs = cos(rs); KFTYPE srs = sin(rs);

	CPose3D  sensorPoseAbs( robotPose + sensorPoseOnRobot );

    //CTicTac   tictac;

	for (i=0;i<nLMs;i++)  // nLMs = 1 if full=false!
	{
		size_t idxOfFeatureInMap;

		if (in_full)
			idxOfFeatureInMap = i;
		else
		{
			ASSERT_(data_association[in_obsIdx]>=0)  // It is an already mapped feature!
			idxOfFeatureInMap = data_association[in_obsIdx];  // Just take one LM
		}

		size_t  row_in = 3*idxOfFeatureInMap;
		size_t  row = 3*i;

		//tictac.Tic();

        // Landmark absolute 3D position in the map:
	    KFTYPE xi = m_xkk[ vehicle_size + row_in + 0 ];
		KFTYPE yi = m_xkk[ vehicle_size + row_in + 1 ];
		KFTYPE zi = m_xkk[ vehicle_size + row_in + 2 ];

		// ---------------------------------------------------
		// Generate Range, yaw, pitch and the ytilde:
		// ---------------------------------------------------
		CPoint3D   mapEst(xi,yi,zi);
		double		h_range, h_yaw,h_pitch;

		sensorPoseAbs.sphericalCoordinates( mapEst, h_range, h_yaw, h_pitch );

		// Fill out "ytilde": If this LM has been observed, the usual value. Otherwise, zeroes:
		int idxThisLMInZ = -1;
		for (size_t p=0;p< data_association.size();p++)
		{
			if ( data_association[p] == static_cast<int>(idxOfFeatureInMap))
			{
				idxThisLMInZ = p;
				break;
			}
		}

		if (idxThisLMInZ<0)
		{
			// Not observed:
			ytilde[row+0] =
			ytilde[row+1] =
			ytilde[row+2] = 0;
		}
		else
		{
			// Yes: compute ytilde:
			ytilde[row+0] = in_z(idxThisLMInZ,0) - h_range;
			ytilde[row+1] = math::wrapToPi( in_z(idxThisLMInZ,1) - h_yaw );
			ytilde[row+2] = math::wrapToPi( in_z(idxThisLMInZ,2) - h_pitch );
		}

        /* //DEBUG
		for (size_t k=0;k<obs->sensedData.size();k++)
            if ((int)obs->sensedData[k].landmarkID== (int)m_IDs_inverse[i] )
            {
                cout << "OBS: " << obs->sensedData[k].range << " " << RAD2DEG(obs->sensedData[k].yaw) << " " << RAD2DEG(obs->sensedData[k].pitch) << endl;
                break;
            }
		cout << "LM: " << mapEst << endl;
		cout << "SensorPose: " << sensorPoseAbs << endl;
		cout << h_range << " " << RAD2DEG(h_yaw) << " " << RAD2DEG(h_pitch) << endl;
		system::pause();
		*/

		// ---------------------------------------------------
		// Generate dhi_dxv: A 6x3 block
		// ---------------------------------------------------

		KFTYPE EXPRE14 = (sy*cr*cys*crs-cy*sp*sr*cys*crs+cy*cp*sys*crs+sy*cr*sys*sps*srs-cps*srs*sy*sr-cy*cp*cys*sps*srs-cps*srs*cy*sp*cr-cy*sp*sr*sys*sps*srs);

		KFTYPE EXPRE17 = EXPRE14*x0+sy*y0*cp*sys*crs+sys*crs*x0s+sps*sp*cys*srs*z0+(sy*cp*cys*sps*srs-sy*cp*sys*crs+sy*sp*sr*sys*sps*srs+sy*sp*sr*cys*crs+cy*cr*sys*sps*srs+cy*cr*cys*crs+cps*srs*sy*sp*cr-cps*srs*cy*sr)*yi+(-sp*cys*sps*srs+sp*sys*crs+cp*sr*sys*sps*srs+cp*sr*cys*crs+cp*cr*cps*srs)*zi-sp*sys*crs*z0-sps*cy*cr*y0*sys*srs-cps*z0s*srs+(cy*cp*cys*sps*srs-cy*cp*sys*crs+cy*sp*sr*sys*sps*srs+cy*sp*sr*cys*crs-sy*cr*sys*sps*srs-sy*cr*cys*crs+cps*srs*cy*sp*cr+cps*srs*sy*sr)*xi+cps*cy*sr*y0*srs-sps*y0s*sys*srs-sps*cys*srs*x0s-sps*sy*y0*cp*cys*srs-sy*sr*y0*sp*cys*crs-sps*sr*cp*sys*srs*z0-sps*sy*sr*y0*sp*sys*srs-cps*cr*cp*srs*z0-cps*sy*cr*y0*sp*srs-sr*cp*cys*crs*z0-y0s*cys*crs-cy*cr*y0*cys*crs;

		KFTYPE EXPRE0 = square(EXPRE17);

		KFTYPE EXPRE18 = (-sy*cr*cys*srs+cy*sp*sr*cys*srs-cy*sp*sr*sys*sps*crs+sy*cr*sys*sps*crs-cps*crs*sy*sr-cps*crs*cy*sp*cr-cy*cp*sys*srs-cy*cp*cys*sps*crs);

		KFTYPE EXPRE35 = (cy*cp*cys*sps*crs+cy*cp*sys*srs+cy*sp*sr*sys*sps*crs-cy*sp*sr*cys*srs-sy*cr*sys*sps*crs+sy*cr*cys*srs+cps*crs*cy*sp*cr+cps*crs*sy*sr);


		KFTYPE EXPRE11 = EXPRE18*x0-cps*z0s*crs-sys*srs*x0s+y0s*cys*srs+sr*cp*cys*srs*z0+cy*cr*y0*cys*srs+cps*cy*sr*y0*crs+sps*sp*cys*crs*z0+EXPRE35*xi+(-sp*cys*sps*crs-sp*sys*srs+cp*sr*sys*sps*crs-cp*sr*cys*srs+cp*cr*cps*crs)*zi+(sy*cp*cys*sps*crs+sy*cp*sys*srs+sy*sp*sr*sys*sps*crs-sy*sp*sr*cys*srs+cy*cr*sys*sps*crs-cy*cr*cys*srs+cps*crs*sy*sp*cr-cps*crs*cy*sr)*yi-sps*sr*cp*sys*crs*z0-sps*sy*sr*y0*sp*sys*crs-sps*cy*cr*y0*sys*crs+sp*sys*srs*z0-sps*sy*y0*cp*cys*crs-cps*cr*cp*crs*z0-sps*y0s*sys*crs-sps*cys*crs*x0s-sy*y0*cp*sys*srs-cps*sy*cr*y0*sp*crs+sy*sr*y0*sp*cys*srs;

		KFTYPE EXPRE20 = (sy*sys*cr*cps-cy*cp*cys*cps+cy*cr*sp*sps-sp*sys*cps*cy*sr+sy*sr*sps);

		KFTYPE EXPRE25 = (-sp*sy*cr*sps+cy*sys*cr*cps+sp*sys*cps*sy*sr+cy*sr*sps+sy*cp*cys*cps);

		KFTYPE EXPRE15 = EXPRE20*x0+(-cy*cr*sp*sps+cy*cp*cys*cps-sy*sr*sps+sp*sys*cps*cy*sr-sy*sys*cr*cps)*xi+EXPRE25*yi+(-sp*cys*cps+cp*sr*sys*cps-cp*cr*sps)*zi-sp*sys*cps*sy*sr*y0+cp*cr*sps*z0-cp*sys*cps*sr*z0+sp*cys*cps*z0+sp*sy*cr*sps*y0-cys*cps*x0s-sys*cps*y0s-sys*cps*cy*cr*y0-cy*sr*sps*y0+z0s*sps-cp*cys*cps*sy*y0;

		KFTYPE EXPRE1 = (0.5)/sqrt(square(EXPRE15)+EXPRE0+square(EXPRE11));
        	KFTYPE EXPRE2 = 2*(( sy*sys*cr*cps-cy*cp*cys*cps+cy*cr*sp*sps-sp*sys*cps*cy*sr+sy*sr*sps)*x0+(-cy*cr*sp*sps+cy*cp*cys*cps-sy*sr*sps+sp*sys*cps*cy*sr-sy*sys*cr*cps)*xi+EXPRE25*yi+(-sp*cys*cps+cp*sr*sys*cps-cp*cr*sps)*zi-sp*sys*cps*sy*sr*y0+cp*cr*sps*z0-cp*sys*cps*sr*z0+sp*cys*cps*z0+sp*sy*cr*sps*y0-cys*cps*x0s-sys*cps*y0s-sys*cps*cy*cr*y0-cy*sr*sps*y0+z0s*sps-cp*cys*cps*sy*y0);
		KFTYPE K_05 = 0.5;
		KFTYPE K_15 = 1.5;

		KFTYPE EXPRE12 = (sy*cp*sys*crs-cy*cr*sys*sps*srs+cps*srs*cy*sr-sy*cp*cys*sps*srs-sy*sp*sr*cys*crs-sy*sp*sr*sys*sps*srs-cps*srs*sy*sp*cr-cy*cr*cys*crs);

		KFTYPE EXPRE13 = ((-sy*sr*cys*crs-cy*sp*cr*cys*crs-sy*sr*sys*sps*srs-cps*srs*sy*cr+cps*srs*cy*sp*sr-cy*sp*cr*sys*sps*srs)*x0+(sy*sp*cr*sys*sps*srs+sy*sp*cr*cys*crs-cy*sr*sys*sps*srs-cy*sr*cys*crs-cps*srs*sy*sp*sr-cps*srs*cy*cr)*yi+(cp*cr*sys*sps*srs+cp*cr*cys*crs-cp*sr*cps*srs)*zi+sps*cy*sr*y0*sys*srs+(cy*sp*cr*sys*sps*srs+cy*sp*cr*cys*crs+sy*sr*sys*sps*srs+sy*sr*cys*crs-cps*srs*cy*sp*sr+cps*srs*sy*cr)*xi+cps*cy*cr*y0*srs-sy*cr*y0*sp*cys*crs-sps*cr*cp*sys*srs*z0-sps*sy*cr*y0*sp*sys*srs+cps*sr*cp*srs*z0+cps*sy*sr*y0*sp*srs-cr*cp*cys*crs*z0+cy*sr*y0*cys*crs);

		KFTYPE EXPRE16 = ((sy*cp*cys*sps*crs+sy*cp*sys*srs+sy*sp*sr*sys*sps*crs-sy*sp*sr*cys*srs+cy*cr*sys*sps*crs-cy*cr*cys*srs+cps*crs*sy*sp*cr-cps*crs*cy*sr)*x0-sy*cr*cys*srs*y0-cps*crs*sy*sr*y0+(cy*cr*cys*srs+cps*crs*cy*sr-sy*sp*sr*sys*sps*crs-cy*cr*sys*sps*crs-sy*cp*cys*sps*crs-sy*cp*sys*srs-cps*crs*sy*sp*cr+sy*sp*sr*cys*srs)*xi+EXPRE35*yi-cy*sp*sr*sys*sps*crs*y0+sy*cr*sys*sps*crs*y0-cy*cp*cys*sps*crs*y0-cy*cp*sys*srs*y0-cps*crs*cy*sp*cr*y0+cy*sp*sr*cys*srs*y0);

		KFTYPE EXPRE21 = (-sp*sys*cps*sy*sr+sp*sy*cr*sps-cy*sys*cr*cps-cy*sr*sps-sy*cp*cys*cps);

		KFTYPE EXPRE22 = (EXPRE25*x0+EXPRE21*xi+(-cy*cr*sp*sps+cy*cp*cys*cps-sy*sr*sps+sp*sys*cps*cy*sr-sy*sys*cr*cps)*yi-sp*sys*cps*cy*sr*y0+sp*cy*cr*sps*y0+sys*cps*sy*cr*y0+sy*sr*sps*y0-cp*cys*cps*cy*y0);

		KFTYPE EXPRE30 = ((cy*sp*cys*cps+cy*cr*cp*sps-cp*sys*cps*cy*sr)*x0+(-cy*cr*cp*sps-cy*sp*cys*cps+cp*sys*cps*cy*sr)*xi+(-cp*sy*cr*sps+cp*sys*cps*sy*sr-sy*sp*cys*cps)*yi+(-cp*cys*cps-sp*sr*sys*cps+sp*cr*sps)*zi-cp*sys*cps*sy*sr*y0-sp*cr*sps*z0+sp*sys*cps*sr*z0+cp*cys*cps*z0+cp*sy*cr*sps*y0+sp*cys*cps*sy*y0);


        Hx(row+0,0)= EXPRE1*(EXPRE2*EXPRE20+2*(EXPRE17)*EXPRE14+2*(EXPRE11)*EXPRE18);
        Hx(row+0,1)= EXPRE1*(EXPRE2*EXPRE21+2*(EXPRE17)*EXPRE12+2*(EXPRE11)*(cy*cr*cys*srs+cps*crs*cy*sr-sy*sp*sr*sys*sps*crs-cy*cr*sys*sps*crs-sy*cp*cys*sps*crs-sy*cp*sys*srs-cps*crs*sy*sp*cr+sy*sp*sr*cys*srs));
        Hx(row+0,2)= EXPRE1*(EXPRE2*(cp*cr*sps-cp*sr*sys*cps+sp*cys*cps)+2*(EXPRE17)*(sp*cys*sps*srs-sp*sys*crs-cp*sr*sys*sps*srs-cp*cr*cps*srs-cp*sr*cys*crs)+2*(EXPRE11)*(cp*sr*cys*srs+sp*cys*sps*crs-cp*sr*sys*sps*crs+sp*sys*srs-cp*cr*cps*crs));
        Hx(row+0,3)= EXPRE1*(EXPRE2*EXPRE22+2*(EXPRE17)*((sy*cp*cys*sps*srs-sy*cp*sys*crs+sy*sp*sr*sys*sps*srs+sy*sp*sr*cys*crs+cy*cr*sys*sps*srs+cy*cr*cys*crs+cps*srs*sy*sp*cr-cps*srs*cy*sr)*x0+cy*cp*sys*crs*y0+(cy*cp*cys*sps*srs-cy*cp*sys*crs+cy*sp*sr*sys*sps*srs+cy*sp*sr*cys*crs-sy*cr*sys*sps*srs-sy*cr*cys*crs+cps*srs*cy*sp*cr+cps*srs*sy*sr)*yi+sy*cr*sys*sps*srs*y0+EXPRE12*xi-cps*srs*sy*sr*y0-cy*cp*cys*sps*srs*y0-sp*cy*sr*cys*crs*y0-sp*cy*sr*sys*sps*srs*y0-cps*srs*cy*cr*sp*y0+sy*cr*cys*crs*y0)+2*(EXPRE11)*EXPRE16);
        Hx(row+0,4)= EXPRE1*(EXPRE2*EXPRE30+2*(EXPRE17)*((-cy*cp*sr*cys*crs-cy*sp*sys*crs+cy*sp*cys*sps*srs-cps*srs*cy*cp*cr-cy*cp*sr*sys*sps*srs)*x0-sy*y0*sp*sys*crs+sps*cp*cys*srs*z0+(-sy*sp*cys*sps*srs+sy*sp*sys*crs+sy*cp*sr*sys*sps*srs+sy*cp*sr*cys*crs+cps*srs*sy*cp*cr)*yi+(-cp*cys*sps*srs+cp*sys*crs-sp*sr*sys*sps*srs-sp*sr*cys*crs-sp*cr*cps*srs)*zi-cp*sys*crs*z0+(-cy*sp*cys*sps*srs+cy*sp*sys*crs+cy*cp*sr*sys*sps*srs+cy*cp*sr*cys*crs+cps*srs*cy*cp*cr)*xi+sps*sy*y0*sp*cys*srs-sy*sr*y0*cp*cys*crs+sps*sr*sp*sys*srs*z0-sps*sy*sr*y0*cp*sys*srs+cps*cr*sp*srs*z0-cps*sy*cr*y0*cp*srs+sr*sp*cys*crs*z0)+2*(EXPRE11)*((cy*cp*sr*cys*srs-cy*cp*sr*sys*sps*crs-cps*crs*cy*cp*cr+cy*sp*sys*srs+cy*sp*cys*sps*crs)*x0-sr*sp*cys*srs*z0+sps*cp*cys*crs*z0+(-cy*sp*cys*sps*crs-cy*sp*sys*srs+cy*cp*sr*sys*sps*crs-cy*cp*sr*cys*srs+cps*crs*cy*cp*cr)*xi+(-cp*cys*sps*crs-cp*sys*srs-sp*sr*sys*sps*crs+sp*sr*cys*srs-sp*cr*cps*crs)*zi+(-sy*sp*cys*sps*crs-sy*sp*sys*srs+sy*cp*sr*sys*sps*crs-sy*cp*sr*cys*srs+cps*crs*sy*cp*cr)*yi+sps*sr*sp*sys*crs*z0-sps*sy*sr*y0*cp*sys*crs+cp*sys*srs*z0+sps*sy*y0*sp*cys*crs+cps*cr*sp*crs*z0+sy*y0*sp*sys*srs-cps*sy*cr*y0*cp*crs+sy*sr*y0*cp*cys*srs));
        Hx(row+0,5)= EXPRE1*(EXPRE2*((-sy*sys*sr*cps-cy*sr*sp*sps-sp*sys*cps*cy*cr+sy*cr*sps)*x0+(cy*sr*sp*sps-sy*cr*sps+sp*sys*cps*cy*cr+sy*sys*sr*cps)*xi+(sp*sy*sr*sps-cy*sys*sr*cps+sp*sys*cps*sy*cr+cy*cr*sps)*yi+(cp*cr*sys*cps+cp*sr*sps)*zi-sp*sys*cps*sy*cr*y0-cp*sr*sps*z0-cp*sys*cps*cr*z0-sp*sy*sr*sps*y0+sys*cps*cy*sr*y0-cy*cr*sps*y0)+2*(EXPRE17)*EXPRE13+2*(EXPRE11)*((sy*sr*cys*srs+cy*sp*cr*cys*srs-cy*sp*cr*sys*sps*crs-sy*sr*sys*sps*crs-cps*crs*sy*cr+cps*crs*cy*sp*sr)*x0+cr*cp*cys*srs*z0-cy*sr*y0*cys*srs+cps*cy*cr*y0*crs+(cy*sp*cr*sys*sps*crs-cy*sp*cr*cys*srs+sy*sr*sys*sps*crs-sy*sr*cys*srs-cps*crs*cy*sp*sr+cps*crs*sy*cr)*xi+(cp*cr*sys*sps*crs-cp*cr*cys*srs-cp*sr*cps*crs)*zi+(sy*sp*cr*sys*sps*crs-sy*sp*cr*cys*srs-cy*sr*sys*sps*crs+cy*sr*cys*srs-cps*crs*sy*sp*sr-cps*crs*cy*cr)*yi-sps*cr*cp*sys*crs*z0-sps*sy*cr*y0*sp*sys*crs+sps*cy*sr*y0*sys*crs+cps*sr*cp*crs*z0+cps*sy*sr*y0*sp*crs+sy*cr*y0*sp*cys*srs));

        Hx(row+1,0)= (EXPRE14/(EXPRE15)-(EXPRE17)/square(EXPRE15)*EXPRE20)/(1+EXPRE0/square(EXPRE15));
        Hx(row+1,1)= (EXPRE12/(EXPRE15)-(EXPRE17)/square(EXPRE15)*EXPRE21)/(1+EXPRE0/square(EXPRE15));
        Hx(row+1,2)= ((sp*cys*sps*srs-sp*sys*crs-cp*sr*sys*sps*srs-cp*cr*cps*srs-cp*sr*cys*crs)/(EXPRE15)-(EXPRE17)/square(EXPRE15)*(cp*cr*sps-cp*sr*sys*cps+sp*cys*cps))/(1+EXPRE0/square(EXPRE15));

        Hx(row+1,3)= (((sy*cp*cys*sps*srs-sy*cp*sys*crs+sy*sp*sr*sys*sps*srs+sy*sp*sr*cys*crs+cy*cr*sys*sps*srs+cy*cr*cys*crs+cps*srs*sy*sp*cr-cps*srs*cy*sr)*x0+cy*cp*sys*crs*y0+(cy*cp*cys*sps*srs-cy*cp*sys*crs+cy*sp*sr*sys*sps*srs+cy*sp*sr*cys*crs-sy*cr*sys*sps*srs-sy*cr*cys*crs+cps*srs*cy*sp*cr+cps*srs*sy*sr)*yi+sy*cr*sys*sps*srs*y0+EXPRE12*xi-cps*srs*sy*sr*y0-cy*cp*cys*sps*srs*y0-sp*cy*sr*cys*crs*y0-sp*cy*sr*sys*sps*srs*y0-cps*srs*cy*cr*sp*y0+sy*cr*cys*crs*y0)/(EXPRE15)-(EXPRE17)/square(EXPRE15)*EXPRE22)/(1+EXPRE0/square(EXPRE15));
        Hx(row+1,4)= (((-cy*cp*sr*cys*crs-cy*sp*sys*crs+cy*sp*cys*sps*srs-cps*srs*cy*cp*cr-cy*cp*sr*sys*sps*srs)*x0-sy*y0*sp*sys*crs+sps*cp*cys*srs*z0+(-sy*sp*cys*sps*srs+sy*sp*sys*crs+sy*cp*sr*sys*sps*srs+sy*cp*sr*cys*crs+cps*srs*sy*cp*cr)*yi+(-cp*cys*sps*srs+cp*sys*crs-sp*sr*sys*sps*srs-sp*sr*cys*crs-sp*cr*cps*srs)*zi-cp*sys*crs*z0+(-cy*sp*cys*sps*srs+cy*sp*sys*crs+cy*cp*sr*sys*sps*srs+cy*cp*sr*cys*crs+cps*srs*cy*cp*cr)*xi+sps*sy*y0*sp*cys*srs-sy*sr*y0*cp*cys*crs+sps*sr*sp*sys*srs*z0-sps*sy*sr*y0*cp*sys*srs+cps*cr*sp*srs*z0-cps*sy*cr*y0*cp*srs+sr*sp*cys*crs*z0)/(EXPRE15)-(EXPRE17)/square(EXPRE15)*EXPRE30)/(1+EXPRE0/square(EXPRE15));
        Hx(row+1,5)= (EXPRE13/(EXPRE15)-(EXPRE17)/square(EXPRE15)*((-sy*sys*sr*cps-cy*sr*sp*sps-sp*sys*cps*cy*cr+sy*cr*sps)*x0+(cy*sr*sp*sps-sy*cr*sps+sp*sys*cps*cy*cr+sy*sys*sr*cps)*xi+(sp*sy*sr*sps-cy*sys*sr*cps+sp*sys*cps*sy*cr+cy*cr*sps)*yi+(cp*cr*sys*cps+cp*sr*sps)*zi-sp*sys*cps*sy*cr*y0-cp*sr*sps*z0-cp*sys*cps*cr*z0-sp*sy*sr*sps*y0+sys*cps*cy*sr*y0-cy*cr*sps*y0))/(1+EXPRE0/square(EXPRE15));

        Hx(row+2,0)= -(EXPRE18/sqrt(square(EXPRE15)+EXPRE0)-K_05*(EXPRE11)/pow(square(EXPRE15)+EXPRE0,K_15)*(EXPRE2*EXPRE20+2*(EXPRE17)*EXPRE14))/(1+square(EXPRE11)/(square(EXPRE15)+EXPRE0));
        Hx(row+2,1)= -((cy*cr*cys*srs+cps*crs*cy*sr-sy*sp*sr*sys*sps*crs-cy*cr*sys*sps*crs-sy*cp*cys*sps*crs-sy*cp*sys*srs-cps*crs*sy*sp*cr+sy*sp*sr*cys*srs)/sqrt(square(EXPRE15)+EXPRE0)-K_05*(EXPRE11)/pow(square(EXPRE15)+EXPRE0,K_15)*(EXPRE2*EXPRE21+2*(EXPRE17)*EXPRE12))/(1+square(EXPRE11)/(square(EXPRE15)+(EXPRE17)));
        Hx(row+2,2)= -((cp*sr*cys*srs+sp*cys*sps*crs-cp*sr*sys*sps*crs+sp*sys*srs-cp*cr*cps*crs)/sqrt(square(EXPRE15)+EXPRE0)-K_05*(EXPRE11)/pow(square(EXPRE15)+EXPRE0,K_15)*(EXPRE2*(cp*cr*sps-cp*sr*sys*cps+sp*cys*cps)+2*(EXPRE17)*(sp*cys*sps*srs-sp*sys*crs-cp*sr*sys*sps*srs-cp*cr*cps*srs-cp*sr*cys*crs)))/(1+square(EXPRE11)/(square(EXPRE15)+EXPRE0));
        Hx(row+2,3)= -(EXPRE16/sqrt(square(EXPRE15)+EXPRE0)-K_05*(EXPRE11)/pow(square(EXPRE15)+EXPRE0,K_15)*(EXPRE2*EXPRE22+2*(EXPRE17)*((sy*cp*cys*sps*srs-sy*cp*sys*crs+sy*sp*sr*sys*sps*srs+sy*sp*sr*cys*crs+cy*cr*sys*sps*srs+cy*cr*cys*crs+cps*srs*sy*sp*cr-cps*srs*cy*sr)*x0+cy*cp*sys*crs*y0+(cy*cp*cys*sps*srs-cy*cp*sys*crs+cy*sp*sr*sys*sps*srs+cy*sp*sr*cys*crs-sy*cr*sys*sps*srs-sy*cr*cys*crs+cps*srs*cy*sp*cr+cps*srs*sy*sr)*yi+sy*cr*sys*sps*srs*y0+EXPRE12*xi-cps*srs*sy*sr*y0-cy*cp*cys*sps*srs*y0-sp*cy*sr*cys*crs*y0-sp*cy*sr*sys*sps*srs*y0-cps*srs*cy*cr*sp*y0+sy*cr*cys*crs*y0)))/(1+square(EXPRE11)/(square(EXPRE15)+EXPRE0));
        Hx(row+2,4)= -(((cy*cp*sr*cys*srs-cy*cp*sr*sys*sps*crs-cps*crs*cy*cp*cr+cy*sp*sys*srs+cy*sp*cys*sps*crs)*x0-sr*sp*cys*srs*z0+sps*cp*cys*crs*z0+(-cy*sp*cys*sps*crs-cy*sp*sys*srs+cy*cp*sr*sys*sps*crs-cy*cp*sr*cys*srs+cps*crs*cy*cp*cr)*xi+(-cp*cys*sps*crs-cp*sys*srs-sp*sr*sys*sps*crs+sp*sr*cys*srs-sp*cr*cps*crs)*zi+(-sy*sp*cys*sps*crs-sy*sp*sys*srs+sy*cp*sr*sys*sps*crs-sy*cp*sr*cys*srs+cps*crs*sy*cp*cr)*yi+sps*sr*sp*sys*crs*z0-sps*sy*sr*y0*cp*sys*crs+cp*sys*srs*z0+sps*sy*y0*sp*cys*crs+cps*cr*sp*crs*z0+sy*y0*sp*sys*srs-cps*sy*cr*y0*cp*crs+sy*sr*y0*cp*cys*srs)/sqrt(square(EXPRE15)+EXPRE0)-K_05*(EXPRE11)/pow(square(EXPRE15)+EXPRE0,K_15)*(EXPRE2*EXPRE30+2*(EXPRE17)*((-cy*cp*sr*cys*crs-cy*sp*sys*crs+cy*sp*cys*sps*srs-cps*srs*cy*cp*cr-cy*cp*sr*sys*sps*srs)*x0-sy*y0*sp*sys*crs+sps*cp*cys*srs*z0+(-sy*sp*cys*sps*srs+sy*sp*sys*crs+sy*cp*sr*sys*sps*srs+sy*cp*sr*cys*crs+cps*srs*sy*cp*cr)*yi+(-cp*cys*sps*srs+cp*sys*crs-sp*sr*sys*sps*srs-sp*sr*cys*crs-sp*cr*cps*srs)*zi-cp*sys*crs*z0+(-cy*sp*cys*sps*srs+cy*sp*sys*crs+cy*cp*sr*sys*sps*srs+cy*cp*sr*cys*crs+cps*srs*cy*cp*cr)*xi+sps*sy*y0*sp*cys*srs-sy*sr*y0*cp*cys*crs+sps*sr*sp*sys*srs*z0-sps*sy*sr*y0*cp*sys*srs+cps*cr*sp*srs*z0-cps*sy*cr*y0*cp*srs+sr*sp*cys*crs*z0)))/(1+square(EXPRE11)/(square(EXPRE15)+EXPRE0));
        Hx(row+2,5)= -(((sy*sr*cys*srs+cy*sp*cr*cys*srs-cy*sp*cr*sys*sps*crs-sy*sr*sys*sps*crs-cps*crs*sy*cr+cps*crs*cy*sp*sr)*x0+cr*cp*cys*srs*z0-cy*sr*y0*cys*srs+cps*cy*cr*y0*crs+(cy*sp*cr*sys*sps*crs-cy*sp*cr*cys*srs+sy*sr*sys*sps*crs-sy*sr*cys*srs-cps*crs*cy*sp*sr+cps*crs*sy*cr)*xi+(cp*cr*sys*sps*crs-cp*cr*cys*srs-cp*sr*cps*crs)*zi+(sy*sp*cr*sys*sps*crs-sy*sp*cr*cys*srs-cy*sr*sys*sps*crs+cy*sr*cys*srs-cps*crs*sy*sp*sr-cps*crs*cy*cr)*yi-sps*cr*cp*sys*crs*z0-sps*sy*cr*y0*sp*sys*crs+sps*cy*sr*y0*sys*crs+cps*sr*cp*crs*z0+cps*sy*sr*y0*sp*crs+sy*cr*y0*sp*cys*srs)/sqrt(square(EXPRE15)+EXPRE0)-K_05*(EXPRE11)/pow(square(EXPRE15)+EXPRE0,K_15)*(EXPRE2*((-sy*sys*sr*cps-cy*sr*sp*sps-sp*sys*cps*cy*cr+sy*cr*sps)*x0+(cy*sr*sp*sps-sy*cr*sps+sp*sys*cps*cy*cr+sy*sys*sr*cps)*xi+(sp*sy*sr*sps-cy*sys*sr*cps+sp*sys*cps*sy*cr+cy*cr*sps)*yi+(cp*cr*sys*cps+cp*sr*sps)*zi-sp*sys*cps*sy*cr*y0-cp*sr*sps*z0-cp*sys*cps*cr*z0-sp*sy*sr*sps*y0+sys*cps*cy*sr*y0-cy*cr*sps*y0)+2*(EXPRE17)*EXPRE13))/(1+square(EXPRE11)/(square(EXPRE15)+EXPRE0));

        // ---------------------------------------------------
        // Generate dhi_dyi: A 3x3 block
        // ---------------------------------------------------
        // Generate dhi_dyi: A 3x3 block
        Hy(row+0,0)=EXPRE1*(EXPRE2*(-cy*cr*sp*sps+cy*cp*cys*cps-sy*sr*sps+sp*sys*cps*cy*sr-sy*sys*cr*cps)+2*(EXPRE17)*(cy*cp*cys*sps*srs-cy*cp*sys*crs+cy*sp*sr*sys*sps*srs+cy*sp*sr*cys*crs-sy*cr*sys*sps*srs-sy*cr*cys*crs+cps*srs*cy*sp*cr+cps*srs*sy*sr)+2*(EXPRE11)*EXPRE35);
        Hy(row+0,1)=EXPRE1*(EXPRE2*EXPRE25+2*(EXPRE17)*(sy*cp*cys*sps*srs-sy*cp*sys*crs+sy*sp*sr*sys*sps*srs+sy*sp*sr*cys*crs+cy*cr*sys*sps*srs+cy*cr*cys*crs+cps*srs*sy*sp*cr-cps*srs*cy*sr)+2*(EXPRE11)*(sy*cp*cys*sps*crs+sy*cp*sys*srs+sy*sp*sr*sys*sps*crs-sy*sp*sr*cys*srs+cy*cr*sys*sps*crs-cy*cr*cys*srs+cps*crs*sy*sp*cr-cps*crs*cy*sr));
        Hy(row+0,2)=EXPRE1*(EXPRE2*(-sp*cys*cps+cp*sr*sys*cps-cp*cr*sps)+2*(EXPRE17)*(-sp*cys*sps*srs+sp*sys*crs+cp*sr*sys*sps*srs+cp*sr*cys*crs+cp*cr*cps*srs)+2*(EXPRE11)*(-sp*cys*sps*crs-sp*sys*srs+cp*sr*sys*sps*crs-cp*sr*cys*srs+cp*cr*cps*crs));

        Hy(row+1,0)=((cy*cp*cys*sps*srs-cy*cp*sys*crs+cy*sp*sr*sys*sps*srs+cy*sp*sr*cys*crs-sy*cr*sys*sps*srs-sy*cr*cys*crs+cps*srs*cy*sp*cr+cps*srs*sy*sr)/(EXPRE15)-(EXPRE17)/square(EXPRE15)*(-cy*cr*sp*sps+cy*cp*cys*cps-sy*sr*sps+sp*sys*cps*cy*sr-sy*sys*cr*cps))/(1+EXPRE0/square(EXPRE15));
        Hy(row+1,1)=((sy*cp*cys*sps*srs-sy*cp*sys*crs+sy*sp*sr*sys*sps*srs+sy*sp*sr*cys*crs+cy*cr*sys*sps*srs+cy*cr*cys*crs+cps*srs*sy*sp*cr-cps*srs*cy*sr)/(EXPRE15)-(EXPRE17)/square(EXPRE15)*EXPRE25)/(1+EXPRE0/square(EXPRE15));
        Hy(row+1,2)=((-sp*cys*sps*srs+sp*sys*crs+cp*sr*sys*sps*srs+cp*sr*cys*crs+cp*cr*cps*srs)/(EXPRE15)-(EXPRE17)/square(EXPRE15)*(-sp*cys*cps+cp*sr*sys*cps-cp*cr*sps))/(1+EXPRE0/square(EXPRE15));

        Hy(row+2,0)=-(EXPRE35/sqrt(square(EXPRE15)+EXPRE0)-K_05*(EXPRE11)/pow(square(EXPRE15)+EXPRE0,K_15)*(EXPRE2*(-cy*cr*sp*sps+cy*cp*cys*cps-sy*sr*sps+sp*sys*cps*cy*sr-sy*sys*cr*cps)+2*(EXPRE17)*(cy*cp*cys*sps*srs-cy*cp*sys*crs+cy*sp*sr*sys*sps*srs+cy*sp*sr*cys*crs-sy*cr*sys*sps*srs-sy*cr*cys*crs+cps*srs*cy*sp*cr+cps*srs*sy*sr)))/(1+square(EXPRE11)/(square(EXPRE15)+EXPRE0));
        Hy(row+2,1)=-((sy*cp*cys*sps*crs+sy*cp*sys*srs+sy*sp*sr*sys*sps*crs-sy*sp*sr*cys*srs+cy*cr*sys*sps*crs-cy*cr*cys*srs+cps*crs*sy*sp*cr-cps*crs*cy*sr)/sqrt(square(EXPRE15)+EXPRE0)-K_05*(EXPRE11)/pow(square(EXPRE15)+EXPRE0,K_15)*(EXPRE2*EXPRE25+2*(EXPRE17)*(sy*cp*cys*sps*srs-sy*cp*sys*crs+sy*sp*sr*sys*sps*srs+sy*sp*sr*cys*crs+cy*cr*sys*sps*srs+cy*cr*cys*crs+cps*srs*sy*sp*cr-cps*srs*cy*sr)))/(1+square(EXPRE11)/(square(EXPRE15)+EXPRE0));
        Hy(row+2,2)=-((-sp*cys*sps*crs-sp*sys*srs+cp*sr*sys*sps*crs-cp*sr*cys*srs+cp*cr*cps*crs)/sqrt(square(EXPRE15)+EXPRE0)-K_05*(EXPRE11)/pow(square(EXPRE15)+EXPRE0,K_15)*(EXPRE2*(-sp*cys*cps+cp*sr*sys*cps-cp*cr*sps)+2*(EXPRE17)*(-sp*cys*sps*srs+sp*sys*crs+cp*sr*sys*sps*srs+cp*sr*cys*crs+cp*cr*cps*srs)))/(1+square(EXPRE11)/(square(EXPRE15)+EXPRE0));

		//printf("LM[%u] takes %f ms\n", (unsigned int)i,tictac.Tac()*1000.0f );
	}

	MRPT_TRY_END
}


/** This is called between the KF prediction step and the update step to allow the application to employ the prior distribution of the system state in the detection of observations (data association), where applicable.
  * \param out_z A \f$N \times O\f$ matrix, with O being get_observation_size() and N the number of "observations": how many observed landmarks for a map, or just one if not applicable.
  * \param out_R A matrix of size N·OxO (O being get_observation_size() and N the number of observations). It is the covariance matrix of the sensor noise for each of the returned observations. The order must correspond to that in out_z.
  * \param out_data_association An empty vector or, where applicable, a vector where the i'th element corresponds to the position of the observation in the i'th row of out_z within the system state vector, or -1 if it is a new map element and we want to insert it at the end of this KF iteration.
  *  This method will be called just once for each complete KF iteration.
  */
void CRangeBearingKFSLAM::OnGetObservations(
	CMatrixTemplateNumeric<KFTYPE> &Z,
	CVectorTemplate<KFTYPE>        &R2,
	vector_int                     &data_association
	)
{
	MRPT_TRY_START

	// Observations: Z
	// ---------------------
	deque<CObservationBearingRange::TMeasurement>::const_iterator itObs;
	vector_int::iterator                                          itDA;

	CObservationBearingRangePtr obs = m_SF->getObservationByClass<CObservationBearingRange>();
	ASSERT_(obs);

	// Z: NxO  (N:#obs, O:3)
	// data_association: N

	size_t N = obs->sensedData.size();
	Z.setSize(N,3);
	data_association.clear();
	data_association.resize(N,-1);

	// For each observed LM:
	size_t row;
    for (row=0,itObs = obs->sensedData.begin(),itDA=data_association.begin();itObs!=obs->sensedData.end();itObs++,itDA++,row++)
    {
    	// Fill one row in Z:
    	Z(row,0) =itObs->range;
    	Z(row,1) =itObs->yaw;
    	Z(row,2) =itObs->pitch;

    	// Fill data asociation:
    	//  Using IDs!
    	std::map<CLandmark::TLandmarkID,unsigned int>::iterator itID;
    	if ( (itID = m_IDs.find( itObs->landmarkID )) != m_IDs.end()  )
    		*itDA = itID->second; // This row in Z corresponds to the i'th map element in the state vector:
    }

	// R2 vector:
	// ---------------------
    R2.resize(3);
    R2[0] = square( options.std_sensor_range );
    R2[1] = square( options.std_sensor_yaw );
    R2[2] = square( options.std_sensor_pitch );

	MRPT_TRY_END

}


/** This virtual function musts normalize the state vector and covariance matrix (only if its necessary).
 */
void  CRangeBearingKFSLAM::OnNormalizeStateVector()
{
	MRPT_TRY_START

	// Check angles:
    m_xkk[3] = math::wrapToPi( m_xkk[3] );
    m_xkk[4] = math::wrapToPi( m_xkk[4] );
    m_xkk[5] = math::wrapToPi( m_xkk[5] );

	MRPT_TRY_END
}

/*---------------------------------------------------------------
						loadOptions
  ---------------------------------------------------------------*/
void CRangeBearingKFSLAM::loadOptions( const mrpt::utils::CConfigFileBase &ini )
{
	// Main
	options.loadFromConfigFile( ini, "RangeBearingKFSLAM" );
	KF_options.loadFromConfigFile( ini, "RangeBearingKFSLAM_KalmanFilter" );
	// partition algorithm:
	mapPartitioner.options.loadFromConfigFile(ini,"RangeBearingKFSLAM");
}

/*---------------------------------------------------------------
						loadOptions
  ---------------------------------------------------------------*/
void CRangeBearingKFSLAM::TOptions::loadFromConfigFile(
	const mrpt::utils::CConfigFileBase	&source,
	const std::string		&section)
{
	stds_Q_no_odo[3] = RAD2DEG(stds_Q_no_odo[3]);
	stds_Q_no_odo[4] = RAD2DEG(stds_Q_no_odo[4]);
	stds_Q_no_odo[5] = RAD2DEG(stds_Q_no_odo[5]);

	source.read_vector(section,"stds_Q_no_odo", stds_Q_no_odo, stds_Q_no_odo );
	ASSERT_(stds_Q_no_odo.size()==6)

	stds_Q_no_odo[3] = DEG2RAD(stds_Q_no_odo[3]);
	stds_Q_no_odo[4] = DEG2RAD(stds_Q_no_odo[4]);
	stds_Q_no_odo[5] = DEG2RAD(stds_Q_no_odo[5]);

	std_sensor_range	= source.read_float(section,"std_sensor_range", std_sensor_range);
	std_sensor_yaw		= DEG2RAD( source.read_float(section,"std_sensor_yaw_deg", RAD2DEG(std_sensor_yaw)));
	std_sensor_pitch	= DEG2RAD( source.read_float(section,"std_sensor_pitch_deg", RAD2DEG(std_sensor_pitch) ));

	std_odo_z_additional = source.read_float(section,"std_odo_z_additional", std_odo_z_additional);

	MRPT_LOAD_CONFIG_VAR(doPartitioningExperiment,bool, source,section);
	MRPT_LOAD_CONFIG_VAR(partitioningMethod,int,  source,section);


	MRPT_LOAD_CONFIG_VAR(quantiles_3D_representation, float, source,section);

}

/*---------------------------------------------------------------
						Constructor
  ---------------------------------------------------------------*/
CRangeBearingKFSLAM::TOptions::TOptions() :
	stds_Q_no_odo(6,0),
	std_sensor_range ( 0.01f),
	std_sensor_yaw   ( DEG2RAD( 0.2f )),
	std_sensor_pitch ( DEG2RAD( 0.2f )),
	std_odo_z_additional ( 0),
	doPartitioningExperiment(false),
	quantiles_3D_representation ( 3),
	partitioningMethod(0)

{
	stds_Q_no_odo[0]=
	stds_Q_no_odo[1]=
	stds_Q_no_odo[2]=0.10f;
	stds_Q_no_odo[3]=
	stds_Q_no_odo[4]=
	stds_Q_no_odo[5]=DEG2RAD(4.0f);
}

/*---------------------------------------------------------------
						dumpToTextStream
  ---------------------------------------------------------------*/
void CRangeBearingKFSLAM::TOptions::dumpToTextStream( CStream &out)
{
	out.printf("\n----------- [CRangeBearingKFSLAM::TOptions] ------------ \n\n");

	//out.printf("stds_Q_no_odo                           = "); stds_Q_no_odo

	out.printf("doPartitioningExperiment                = %c\n",	doPartitioningExperiment ? 'Y':'N' );
	out.printf("partitioningMethod                      = %i\n",	partitioningMethod );

}

/** If applicable to the given problem, this method implements the inverse observation model needed to extend the "map" with a new "element".
* \param in_z This is the same matrix returned by OnGetObservations().
* \param in_obsIndex The index of the observation whose inverse sensor is to be computed. It corresponds to the row in in_z where the observation can be found.
* \param in_idxNewFeat The index that this new feature will have in the state vector (0:just after the vehicle state, 1: after that,...). Save this number so data association can be done according to these indices.
* \param out_yn The F-length vector with the inverse observation model \f$ y_n=y(x,z_n) \f$.
* \param out_dyn_dxv The \f$F \times V\f$ Jacobian of the inv. sensor model wrt the robot pose \f$ \frac{\partial y_n}{\partial x_v} \f$.
* \param out_dyn_dhn The \f$F \times O\f$ Jacobian of the inv. sensor model wrt the observation vector \f$ \frac{\partial y_n}{\partial h_n} \f$.
*
*  - O: get_observation_size()
*  - V: get_vehicle_size()
*  - F: get_feature_size()
*/
void  CRangeBearingKFSLAM::OnInverseObservationModel(
	const CMatrixTemplateNumeric<KFTYPE> &in_z,
	const size_t                         &in_obsIdx,
	const size_t                         &in_idxNewFeat,
	CVectorTemplate<KFTYPE>              &yn,
	CMatrixTemplateNumeric<KFTYPE>       &dyn_dxv,
	CMatrixTemplateNumeric<KFTYPE>       &dyn_dhn )
{
	MRPT_TRY_START

	/* -------------------------------------------
       Equations, obtained using matlab
        Refer to technical report "6D EKF derivation...", 2008

        x0 y0 z0 y p r         % Robot's 6D pose
        x0s y0s z0s ys ps rs   % Sensor's 6D pose relative to robot
        hn_r hn_y hn_p         % Observation hn: range, yaw, pitch

        xn yn zn               % Absolute 3D landmark coordinates:

        dyn_dxv   -> Jacobian of the inv. observation model wrt the robot pose
        dyn_dhn   -> Jacobian of the inv. observation model wrt each landmark observation

        Sizes:
	     hn:      1x3  <--
         yn:      1x3  -->
         dyn_dxv: 3x6  -->
         dyn_dhn: 3x3  -->
	  ------------------------------------------- */

	CObservationBearingRangePtr obs = m_SF->getObservationByClass<CObservationBearingRange>();
	ASSERT_(obs);
	CPose3D   sensorPoseOnRobot( obs->sensorLocationOnRobot );

	// Mean of the prior of the robot pose:
	CPose3D  robotPose( m_xkk[0],m_xkk[1],m_xkk[2],m_xkk[3],m_xkk[4],m_xkk[5] );

	yn.resize(3);
    dyn_dxv.setSize(3,6);
    dyn_dhn.setSize(3,3);

    // Robot 6D pose:
    KFTYPE x0 = m_xkk[0];
    KFTYPE y0 = m_xkk[1];
    KFTYPE z0 = m_xkk[2];
    KFTYPE y  = m_xkk[3];
    KFTYPE p  = m_xkk[4];
    KFTYPE r  = m_xkk[5];

    KFTYPE cy = cos(y); KFTYPE sy = sin(y);
    KFTYPE cp = cos(p); KFTYPE sp = sin(p);
    KFTYPE cr = cos(r); KFTYPE sr = sin(r);

    // Sensor 6D pose on robot:
    KFTYPE x0s = sensorPoseOnRobot.x;
    KFTYPE y0s = sensorPoseOnRobot.y;
    KFTYPE z0s = sensorPoseOnRobot.z;
    KFTYPE ys = sensorPoseOnRobot.yaw;
    KFTYPE ps = sensorPoseOnRobot.pitch;
    KFTYPE rs = sensorPoseOnRobot.roll;

    KFTYPE cys = cos(ys); KFTYPE sys = sin(ys);
    KFTYPE cps = cos(ps); KFTYPE sps = sin(ps);
    KFTYPE crs = cos(rs); KFTYPE srs = sin(rs);

    ASSERT_(in_z.getColCount()==3)
    KFTYPE hn_r = in_z(in_obsIdx,0);
    KFTYPE hn_y = in_z(in_obsIdx,1);
    KFTYPE hn_p = in_z(in_obsIdx,2);

    KFTYPE chn_y = cos(hn_y);  KFTYPE shn_y = sin(hn_y);
    KFTYPE chn_p = cos(hn_p);  KFTYPE shn_p = sin(hn_p);

    // Compute the mean 3D absolute coordinates:
    yn[0] = (cy*cp*cys*cps+(cy*sp*sr-sy*cr)*sys*cps-(cy*sp*cr+sy*sr)*sps)*hn_r*chn_y*chn_p+(cy*cp*(cys*sps*srs-sys*crs)+(cy*sp*sr-sy*cr)*(sys*sps*srs+cys*crs)+(cy*sp*cr+sy*sr)*cps*srs)*hn_r*shn_y*chn_p-(cy*cp*(cys*sps*crs+sys*srs)+(cy*sp*sr-sy*cr)*(sys*sps*crs-cys*srs)+(cy*sp*cr+sy*sr)*cps*crs)*hn_r*shn_p+cy*cp*x0s+(cy*sp*sr-sy*cr)*y0s+(cy*sp*cr+sy*sr)*z0s+x0;
    yn[1] = (sy*cp*cys*cps+(sy*sp*sr+cy*cr)*sys*cps-(sy*sp*cr-cy*sr)*sps)*hn_r*chn_y*chn_p+(sy*cp*(cys*sps*srs-sys*crs)+(sy*sp*sr+cy*cr)*(sys*sps*srs+cys*crs)+(sy*sp*cr-cy*sr)*cps*srs)*hn_r*shn_y*chn_p-(sy*cp*(cys*sps*crs+sys*srs)+(sy*sp*sr+cy*cr)*(sys*sps*crs-cys*srs)+(sy*sp*cr-cy*sr)*cps*crs)*hn_r*shn_p+sy*cp*x0s+(sy*sp*sr+cy*cr)*y0s+(sy*sp*cr-cy*sr)*z0s+y0;
    yn[2] = (-sp*cys*cps+cp*sr*sys*cps-cp*cr*sps)*hn_r*chn_y*chn_p+(-sp*(cys*sps*srs-sys*crs)+cp*sr*(sys*sps*srs+cys*crs)+cp*cr*cps*srs)*hn_r*shn_y*chn_p-(-sp*(cys*sps*crs+sys*srs)+cp*sr*(sys*sps*crs-cys*srs)+cp*cr*cps*crs)*hn_r*shn_p-sp*x0s+cp*sr*y0s+cp*cr*z0s+z0;

    // Jacobian wrt xv:
    dyn_dxv(0,0) = 1;
    dyn_dxv(0,1) = 0;
    dyn_dxv(0,2) = 0;
    dyn_dxv(0,3) = (-sy*cp*cys*cps+(-sy*sp*sr-cy*cr)*sys*cps-(-sy*sp*cr+cy*sr)*sps)*hn_r*chn_y*chn_p+(-sy*cp*(cys*sps*srs-sys*crs)+(-sy*sp*sr-cy*cr)*(sys*sps*srs+cys*crs)+(-sy*sp*cr+cy*sr)*cps*srs)*hn_r*shn_y*chn_p-(-sy*cp*(cys*sps*crs+sys*srs)+(-sy*sp*sr-cy*cr)*(sys*sps*crs-cys*srs)+(-sy*sp*cr+cy*sr)*cps*crs)*hn_r*shn_p-sy*cp*x0s+(-sy*sp*sr-cy*cr)*y0s+(-sy*sp*cr+cy*sr)*z0s;
    dyn_dxv(0,4) = (-cy*sp*cys*cps+cy*cp*sr*sys*cps-cy*cp*cr*sps)*hn_r*chn_y*chn_p+(-cy*sp*(cys*sps*srs-sys*crs)+cy*cp*sr*(sys*sps*srs+cys*crs)+cy*cp*cr*cps*srs)*hn_r*shn_y*chn_p-(-cy*sp*(cys*sps*crs+sys*srs)+cy*cp*sr*(sys*sps*crs-cys*srs)+cy*cp*cr*cps*crs)*hn_r*shn_p-cy*sp*x0s+cy*cp*sr*y0s+cy*cp*cr*z0s;
    dyn_dxv(0,5) = ((cy*sp*cr+sy*sr)*sys*cps-(-cy*sp*sr+sy*cr)*sps)*hn_r*chn_y*chn_p+((cy*sp*cr+sy*sr)*(sys*sps*srs+cys*crs)+(-cy*sp*sr+sy*cr)*cps*srs)*hn_r*shn_y*chn_p-((cy*sp*cr+sy*sr)*(sys*sps*crs-cys*srs)+(-cy*sp*sr+sy*cr)*cps*crs)*hn_r*shn_p+(cy*sp*cr+sy*sr)*y0s+(-cy*sp*sr+sy*cr)*z0s;

    dyn_dxv(1,0) = 0;
    dyn_dxv(1,1) = 1;
    dyn_dxv(1,2) = 0;
    dyn_dxv(1,3) = (cy*cp*cys*cps+(cy*sp*sr-sy*cr)*sys*cps-(cy*sp*cr+sy*sr)*sps)*hn_r*chn_y*chn_p+(cy*cp*(cys*sps*srs-sys*crs)+(cy*sp*sr-sy*cr)*(sys*sps*srs+cys*crs)+(cy*sp*cr+sy*sr)*cps*srs)*hn_r*shn_y*chn_p-(cy*cp*(cys*sps*crs+sys*srs)+(cy*sp*sr-sy*cr)*(sys*sps*crs-cys*srs)+(cy*sp*cr+sy*sr)*cps*crs)*hn_r*shn_p+cy*cp*x0s+(cy*sp*sr-sy*cr)*y0s+(cy*sp*cr+sy*sr)*z0s;
    dyn_dxv(1,4) = (-sy*sp*cys*cps+sy*cp*sr*sys*cps-sy*cp*cr*sps)*hn_r*chn_y*chn_p+(-sy*sp*(cys*sps*srs-sys*crs)+sy*cp*sr*(sys*sps*srs+cys*crs)+sy*cp*cr*cps*srs)*hn_r*shn_y*chn_p-(-sy*sp*(cys*sps*crs+sys*srs)+sy*cp*sr*(sys*sps*crs-cys*srs)+sy*cp*cr*cps*crs)*hn_r*shn_p-sy*sp*x0s+sy*cp*sr*y0s+sy*cp*cr*z0s;
    dyn_dxv(1,5) = ((sy*sp*cr-cy*sr)*sys*cps-(-sy*sp*sr-cy*cr)*sps)*hn_r*chn_y*chn_p+((sy*sp*cr-cy*sr)*(sys*sps*srs+cys*crs)+(-sy*sp*sr-cy*cr)*cps*srs)*hn_r*shn_y*chn_p-((sy*sp*cr-cy*sr)*(sys*sps*crs-cys*srs)+(-sy*sp*sr-cy*cr)*cps*crs)*hn_r*shn_p+(sy*sp*cr-cy*sr)*y0s+(-sy*sp*sr-cy*cr)*z0s;

    dyn_dxv(2,0) = 0;
    dyn_dxv(2,1) = 0;
    dyn_dxv(2,2) = 1;
    dyn_dxv(2,3) = 0;
    dyn_dxv(2,4) = (-cp*cys*cps-sp*sr*sys*cps+sp*cr*sps)*hn_r*chn_y*chn_p+(-cp*(cys*sps*srs-sys*crs)-sp*sr*(sys*sps*srs+cys*crs)-sp*cr*cps*srs)*hn_r*shn_y*chn_p-(-cp*(cys*sps*crs+sys*srs)-sp*sr*(sys*sps*crs-cys*srs)-sp*cr*cps*crs)*hn_r*shn_p-cp*x0s-sp*sr*y0s-sp*cr*z0s;
    dyn_dxv(2,5) = (cp*cr*sys*cps+cp*sr*sps)*hn_r*chn_y*chn_p+(cp*cr*(sys*sps*srs+cys*crs)-cp*sr*cps*srs)*hn_r*shn_y*chn_p-(cp*cr*(sys*sps*crs-cys*srs)-cp*sr*cps*crs)*hn_r*shn_p+cp*cr*y0s-cp*sr*z0s;

    // Jacobian wrt hn:
    dyn_dhn(0,0) = (cy*cp*cys*cps+(cy*sp*sr-sy*cr)*sys*cps-(cy*sp*cr+sy*sr)*sps)*chn_y*chn_p+(cy*cp*(cys*sps*srs-sys*crs)+(cy*sp*sr-sy*cr)*(sys*sps*srs+cys*crs)+(cy*sp*cr+sy*sr)*cps*srs)*shn_y*chn_p-(cy*cp*(cys*sps*crs+sys*srs)+(cy*sp*sr-sy*cr)*(sys*sps*crs-cys*srs)+(cy*sp*cr+sy*sr)*cps*crs)*shn_p;
    dyn_dhn(0,1) = -(cy*cp*cys*cps+(cy*sp*sr-sy*cr)*sys*cps-(cy*sp*cr+sy*sr)*sps)*hn_r*shn_y*chn_p+(cy*cp*(cys*sps*srs-sys*crs)+(cy*sp*sr-sy*cr)*(sys*sps*srs+cys*crs)+(cy*sp*cr+sy*sr)*cps*srs)*hn_r*chn_y*chn_p;
    dyn_dhn(0,2) = -(cy*cp*cys*cps+(cy*sp*sr-sy*cr)*sys*cps-(cy*sp*cr+sy*sr)*sps)*hn_r*chn_y*shn_p-(cy*cp*(cys*sps*srs-sys*crs)+(cy*sp*sr-sy*cr)*(sys*sps*srs+cys*crs)+(cy*sp*cr+sy*sr)*cps*srs)*hn_r*shn_y*shn_p-(cy*cp*(cys*sps*crs+sys*srs)+(cy*sp*sr-sy*cr)*(sys*sps*crs-cys*srs)+(cy*sp*cr+sy*sr)*cps*crs)*hn_r*chn_p;

    dyn_dhn(1,0) = (sy*cp*cys*cps+(sy*sp*sr+cy*cr)*sys*cps-(sy*sp*cr-cy*sr)*sps)*chn_y*chn_p+(sy*cp*(cys*sps*srs-sys*crs)+(sy*sp*sr+cy*cr)*(sys*sps*srs+cys*crs)+(sy*sp*cr-cy*sr)*cps*srs)*shn_y*chn_p-(sy*cp*(cys*sps*crs+sys*srs)+(sy*sp*sr+cy*cr)*(sys*sps*crs-cys*srs)+(sy*sp*cr-cy*sr)*cps*crs)*shn_p;
    dyn_dhn(1,1) =-(sy*cp*cys*cps+(sy*sp*sr+cy*cr)*sys*cps-(sy*sp*cr-cy*sr)*sps)*hn_r*shn_y*chn_p+(sy*cp*(cys*sps*srs-sys*crs)+(sy*sp*sr+cy*cr)*(sys*sps*srs+cys*crs)+(sy*sp*cr-cy*sr)*cps*srs)*hn_r*chn_y*chn_p;
    dyn_dhn(1,2) = -(sy*cp*cys*cps+(sy*sp*sr+cy*cr)*sys*cps-(sy*sp*cr-cy*sr)*sps)*hn_r*chn_y*shn_p-(sy*cp*(cys*sps*srs-sys*crs)+(sy*sp*sr+cy*cr)*(sys*sps*srs+cys*crs)+(sy*sp*cr-cy*sr)*cps*srs)*hn_r*shn_y*shn_p-(sy*cp*(cys*sps*crs+sys*srs)+(sy*sp*sr+cy*cr)*(sys*sps*crs-cys*srs)+(sy*sp*cr-cy*sr)*cps*crs)*hn_r*chn_p;

    dyn_dhn(2,0) =(-sp*cys*cps+cp*sr*sys*cps-cp*cr*sps)*chn_y*chn_p+(-sp*(cys*sps*srs-sys*crs)+cp*sr*(sys*sps*srs+cys*crs)+cp*cr*cps*srs)*shn_y*chn_p-(-sp*(cys*sps*crs+sys*srs)+cp*sr*(sys*sps*crs-cys*srs)+cp*cr*cps*crs)*shn_p;
    dyn_dhn(2,1) =-(-sp*cys*cps+cp*sr*sys*cps-cp*cr*sps)*hn_r*shn_y*chn_p+(-sp*(cys*sps*srs-sys*crs)+cp*sr*(sys*sps*srs+cys*crs)+cp*cr*cps*srs)*hn_r*chn_y*chn_p;
    dyn_dhn(2,2) =-(-sp*cys*cps+cp*sr*sys*cps-cp*cr*sps)*hn_r*chn_y*shn_p-(-sp*(cys*sps*srs-sys*crs)+cp*sr*(sys*sps*srs+cys*crs)+cp*cr*cps*srs)*hn_r*shn_y*shn_p-(-sp*(cys*sps*crs+sys*srs)+cp*sr*(sys*sps*crs-cys*srs)+cp*cr*cps*crs)*hn_r*chn_p;



	// ----------------------------------------------
	// introduce in the lists of ID<->index in map:
	// ----------------------------------------------
	ASSERT_( in_obsIdx < obs->sensedData.size() )
	ASSERT_( obs->sensedData[in_obsIdx].landmarkID>=0 )

	m_IDs[ obs->sensedData[in_obsIdx].landmarkID ] = in_idxNewFeat;
	m_IDs_inverse[ in_idxNewFeat ] = obs->sensedData[in_obsIdx].landmarkID;


    MRPT_TRY_END
}


/*---------------------------------------------------------------
						getAs3DObject
  ---------------------------------------------------------------*/
void  CRangeBearingKFSLAM::getAs3DObject( mrpt::opengl::CSetOfObjectsPtr	&outObj ) const
{
    outObj->clear();

	// ------------------------------------------------
	//  Add the XYZ corner for the current area:
	// ------------------------------------------------
	outObj->insert( opengl::stock_objects::CornerXYZ() );


	// 3D ellipsoid for robot pose:
	CPointPDFGaussian	pointGauss;
    pointGauss.mean.x = m_xkk[0];
    pointGauss.mean.y = m_xkk[1];
    pointGauss.mean.z = m_xkk[2];
    CMatrixTemplateNumeric<KFTYPE>  COV;
    m_pkk.extractMatrix(0,0,3,3, COV);
    pointGauss.cov = COV;

    {
		opengl::CEllipsoidPtr ellip = opengl::CEllipsoid::Create();

		ellip->setPose(pointGauss.mean);
		ellip->setCovMatrix(pointGauss.cov);
		ellip->enableDrawSolid3D(false);
		ellip->setQuantiles( options.quantiles_3D_representation );
		ellip->set3DsegmentsCount(10);
		ellip->setColor(1,0,0);

		outObj->insert( ellip );
    }


	// 3D ellipsoids for landmarks:
	size_t i,nLMs = (m_xkk.size()-6)/3;
	for (i=0;i<nLMs;i++)
	{
        pointGauss.mean.x = m_xkk[6+3*i+0];
        pointGauss.mean.y = m_xkk[6+3*i+1];
        pointGauss.mean.z = m_xkk[6+3*i+2];
        m_pkk.extractMatrix(6+3*i,6+3*i,3,3, COV);
        pointGauss.cov = COV;

		opengl::CEllipsoidPtr ellip = opengl::CEllipsoid::Create();

		ellip->setName( format( "%"PRIuPTR,i ) );
		ellip->enableShowName(true);
		ellip->setPose( pointGauss.mean );
		ellip->setCovMatrix( pointGauss.cov );
		ellip->enableDrawSolid3D(false);
		ellip->setQuantiles( options.quantiles_3D_representation );
		ellip->set3DsegmentsCount(10);

		ellip->setColor(0,0,1);

		// Set color depending on partitions?
		if (options.doPartitioningExperiment)
		{
		    // This is done for each landmark:
			map<int,bool>    belongToPartition;
			const CSensFrameProbSequence *SFs = &m_SFs; //mapPartitioner.getSequenceOfFrames();

			for (size_t p=0;p<m_lastPartitionSet.size();p++)
			{
				for (size_t w=0;w<m_lastPartitionSet[p].size();w++)
				{
					// Check if landmark #i is in the SF of m_lastPartitionSet[p][w]:
					CLandmark::TLandmarkID  i_th_ID = INVALID_LANDMARK_ID;
					for (map<CLandmark::TLandmarkID,unsigned int>::const_iterator it=m_IDs.begin();it!=m_IDs.end();it++)
					{
						if (it->second==i)
						{
							i_th_ID = it->first;
							break;
						}
					}

					ASSERT_( i_th_ID != INVALID_LANDMARK_ID);

					// Look for the lm_ID in the SF:
					CPose3DPDFPtr    pdf;
					CSensoryFramePtr SF_i;
					SFs->get( m_lastPartitionSet[p][w], pdf, SF_i);

					CObservationBearingRangePtr obs = SF_i->getObservationByClass<CObservationBearingRange>();

					for (size_t o=0;o<obs->sensedData.size();o++)
					{
						if ( obs->sensedData[o].landmarkID == i_th_ID )
						{
							belongToPartition[p]=true;
							break;
						}
					} // end for o
				} // end for w
			} // end for p

			// Build label:
			string strParts("[");

			for (map<int,bool>::iterator it=belongToPartition.begin();it!=belongToPartition.end();it++)
			{
				if (it!=belongToPartition.begin()) strParts += string(",");
				strParts += format("%i",it->first);
			}

			ellip->setName( ellip->getName() + strParts + string("]") );

		}

		outObj->insert( ellip );
	}
}


/*---------------------------------------------------------------
						getLastPartitionLandmarks
  ---------------------------------------------------------------*/
void CRangeBearingKFSLAM::getLastPartitionLandmarksAsIfFixedSubmaps( size_t K, std::vector<vector_uint>	&landmarksMembership )
{
    // Temporal copy:
    std::vector<vector_uint> tmpParts = m_lastPartitionSet;

    // Fake "m_lastPartitionSet":

    // Fixed partitions every K observations:
    vector<vector_uint>   partitions;
    vector_uint           tmpCluster;

    for (size_t i=0;i<m_SFs.size();i++)
    {
        tmpCluster.push_back(i);
        if ( (i % K) == 0 )
        {
            partitions.push_back(tmpCluster);
            tmpCluster.clear();
            tmpCluster.push_back(i);  // This observation "i" is shared between both clusters
        }
    }
    m_lastPartitionSet = partitions;

    // Call the actual method:
    getLastPartitionLandmarks(landmarksMembership);

    // Replace copy:
    m_lastPartitionSet = tmpParts;
}

/*---------------------------------------------------------------
						getLastPartitionLandmarks
  ---------------------------------------------------------------*/
void CRangeBearingKFSLAM::getLastPartitionLandmarks(
    std::vector<vector_uint>	&landmarksMembership ) const
{
    landmarksMembership.clear();

    // All the computation is made based on "m_lastPartitionSet"

    if (!options.doPartitioningExperiment) return;

	size_t i,nLMs = (m_xkk.size()-6)/3;
	for (i=0;i<nLMs;i++)
	{
        map<int,bool>    belongToPartition;
        const CSensFrameProbSequence *SFs = &m_SFs; //mapPartitioner.getSequenceOfFrames();

        for (size_t p=0;p<m_lastPartitionSet.size();p++)
        {
            for (size_t w=0;w<m_lastPartitionSet[p].size();w++)
            {
                // Check if landmark #i is in the SF of m_lastPartitionSet[p][w]:
                CLandmark::TLandmarkID  i_th_ID = INVALID_LANDMARK_ID;
                for (map<CLandmark::TLandmarkID,unsigned int>::const_iterator it=m_IDs.begin();it!=m_IDs.end();it++)
                {
                    if (it->second==i)
                    {
                        i_th_ID = it->first;
                        break;
                    }
                }

                ASSERT_( i_th_ID != INVALID_LANDMARK_ID);

                // Look for the lm_ID in the SF:
                CPose3DPDFPtr pdf;
                CSensoryFramePtr SF_i;
                SFs->get( m_lastPartitionSet[p][w], pdf, SF_i);

                CObservationBearingRangePtr obs = SF_i->getObservationByClass<CObservationBearingRange>();

                for (size_t o=0;o<obs->sensedData.size();o++)
                {
                    if ( obs->sensedData[o].landmarkID == i_th_ID )
                    {
                        belongToPartition[p]=true;
                        break;
                    }
                } // end for o
            } // end for w
        } // end for p

        // Build membership list:
        vector_uint membershipOfThisLM;

        for (map<int,bool>::iterator it=belongToPartition.begin();it!=belongToPartition.end();it++)
            membershipOfThisLM.push_back(it->first);

        landmarksMembership.push_back( membershipOfThisLM );
	} // end for i
}


/*---------------------------------------------------------------
            computeOffDiagonalBlocksApproximationError
  ---------------------------------------------------------------*/
double CRangeBearingKFSLAM::computeOffDiagonalBlocksApproximationError(
    const std::vector<vector_uint>	&landmarksMembership ) const
{
    MRPT_TRY_START

    // Compute the information matrix:
    CMatrixTemplateNumeric<KFTYPE> fullCov( m_pkk );
	size_t i;
    for (i=0;i<6;i++)
        fullCov(i,i) = max(fullCov(i,i), 1e-6);

    CMatrixTemplateNumeric<KFTYPE>		H( fullCov.inv() );
    H.Abs();        // Replace by absolute values:

    double sumOffBlocks = 0;
    unsigned int nLMs = landmarksMembership.size();

    ASSERT_((6+nLMs*3)==fullCov.getColCount());

    for (i=0;i<nLMs;i++)
    {
        for (size_t j=i+1;j<nLMs;j++)
        {
            // Sum the cross cov. between LM(i) and LM(j)??
            // --> Only if there is no common cluster:
            if ( 0==math::countCommonElements( landmarksMembership[i],landmarksMembership[j] ) )
            {
                size_t col = 6+i*3;
                size_t row = 6+j*3;
                sumOffBlocks += 2 * H.sum( row,col,row+2,col+2 );
            }
        }
    }

    return sumOffBlocks / H.sum(6,6);  // Starting (6,6)-end
    MRPT_TRY_END
}

/*---------------------------------------------------------------
              reconsiderPartitionsNow
  ---------------------------------------------------------------*/
void CRangeBearingKFSLAM::reconsiderPartitionsNow()
{
	mapPartitioner.markAllNodesForReconsideration();

    vector<vector_uint>   partitions; // A different buffer for making this thread-safe some day...

    mapPartitioner.updatePartitions( partitions );

    m_lastPartitionSet = partitions;
}

/*---------------------------------------------------------------
              saveMapAndPathRepresentationAsMATLABFile
  ---------------------------------------------------------------*/
void CRangeBearingKFSLAM::saveMapAndPath2DRepresentationAsMATLABFile(
	const string &fil,
	float stdCount,
	const string &styleLandmarks,
	const string &stylePath,
	const string &styleRobot ) const
{
	FILE	*f= os::fopen(fil.c_str(),"wt");
	if (!f) return;

	CMatrixDouble cov(2,2);
	CVectorDouble mean(2);

	// Header:
	os::fprintf(f,"%%--------------------------------------------------------------------\n");
	os::fprintf(f,"%% File automatically generated using the MRPT method:\n");
	os::fprintf(f,"%% 'CRangeBearingKFSLAM::saveMapAndPath2DRepresentationAsMATLABFile'\n");
	os::fprintf(f,"%%\n");
	os::fprintf(f,"%%                        ~ MRPT ~\n");
	os::fprintf(f,"%%  Jose Luis Blanco Claraco, University of Malaga @ 2008\n");
	os::fprintf(f,"%%  http://mrpt.sourceforge.net/ \n");
	os::fprintf(f,"%%--------------------------------------------------------------------\n");

	// Main code:
	os::fprintf(f,"hold on;\n\n");

	size_t i, nLMs = (m_xkk.size()-get_vehicle_size())/get_feature_size();

	for (i=0;i<nLMs;i++)
	{
		size_t idx = 6+i*3;

		cov(0,0) = m_pkk(idx+0,idx+0);
		cov(1,1) = m_pkk(idx+1,idx+1);
		cov(0,1) = cov(1,0) = m_pkk(idx+0,idx+1);

		mean[0] = m_xkk[idx+0];
		mean[1] = m_xkk[idx+1];

		// Command to draw the 2D ellipse:
		os::fprintf(f, "%s", math::MATLAB_plotCovariance2D( cov, mean, stdCount,styleLandmarks ).c_str() );
	}


	// Now: the robot path:
	// ------------------------------
	if (m_SFs.size())
	{
		os::fprintf(f,"\nROB_PATH=[");
		for (i=0;i<m_SFs.size();i++)
		{
			CSensoryFramePtr dummySF;
			CPose3DPDFPtr pdf3D;
			m_SFs.get(i,pdf3D,dummySF);

			CPoint3D  pnt3D( pdf3D->getEstimatedPose() ); // 6D -> 3D only

			os::fprintf(f,"%.04f %.04f", pnt3D.x, pnt3D.y );
			if (i<(m_SFs.size()-1))
				os::fprintf(f,";");
		}
		os::fprintf(f,"];\n");

		os::fprintf(f,"plot(ROB_PATH(:,1),ROB_PATH(:,2),'%s');\n",stylePath.c_str());
	}

	// The robot pose:
	cov(0,0) = m_pkk(0,0);
	cov(1,1) = m_pkk(1,1);
	cov(0,1) = cov(1,0) = m_pkk(0,1);

	mean[0] = m_xkk[0];
	mean[1] = m_xkk[1];

	os::fprintf(f, "%s", math::MATLAB_plotCovariance2D( cov, mean, stdCount, styleRobot ).c_str() );

	os::fprintf(f,"\naxis equal;\n");
	os::fclose(f);
}
